"""
This module provides functionality for post-translational modifications.

Wraps modifications in a structured class and allows filtering of
modifications by amino acid and modification type.
"""

LABEL_NAMES = ["TMT", "ITRAQ"]


class Modifications:
    """
    A list of modifications.

    Wraps the Modification objects and provides several utility functions.

    Attributes
    ----------
    mods : list of :class:`Modification<pyproteome.modification.Modification>`
    """

    def __init__(self, mods):
        """
        Initialize from a list of modifications.

        Parameters
        ----------
        mods : list of :class:`Modification<pyproteome.modification.Modification>`
        """
        self.mods = mods

    def __iter__(self):
        return iter(self.mods)

    def __hash__(self):
        return hash(
            tuple(self.mods),
        )

    def __len__(self):
        return len(self.mods)

    def skip_labels_iter(self):
        """
        Return an iterable, skipping over any modifications for peptide labels.

        Returns
        -------
        generator of :class:`Modification<pyproteome.modification.Modification>`
        """
        return (
            mod
            for mod in self.mods
            if not any(label in mod.mod_type for label in LABEL_NAMES)
        )

    def get_mods(self, letter_mod_types):
        """
        Filter the list of modifications.

        Only keeps modifications with a given letter, mod_type, or both.

        Parameters
        ----------
        letter_mod_types : list of tuple of str, str

        Returns
        -------
        generator of Modification
        """
        any_letter, any_mod, letter_mod = \
            _extract_letter_mods(letter_mod_types)
        return (
            mod
            for mod in self.mods
            if allowed_mod_type(
                mod,
                any_letter=any_letter,
                any_mod=any_mod,
                letter_mod=letter_mod,
            )
        )

    def __eq__(self, other):
        if not isinstance(other, Modifications):
            raise TypeError()

        # XXX: Incorrect if modifications are sorted differently
        return len(self.mods) == len(other.mods) and all(
            i == j
            for i, j in zip(self.mods, other.mods)
        )

    def __str__(self, absolute=True, skip_labels=True):
        if len(self.mods) == 0:
            return ""

        if skip_labels:
            lst = list(self.skip_labels_iter())
        else:
            lst = list(iter(self))

        return " / ".join(
            ", ".join(
                "{}{}{}{}".format(
                    mod.display_mod_type(),
                    mod.letter.upper(),
                    1 + (mod.abs_pos[i] if absolute else mod.rel_pos),
                    "" if mod.exact[i] else "*"
                )
                for mod in lst
            )
            for i in range(len(self.mods[0].exact))
            if lst
        )


class Modification:
    """
    Contains information for a single peptide modification.

    Attributes
    ----------
    rel_pos : int
    mod_type : str
    nterm : bool
    cterm : bool
    letter : str
    abs_pos : int
    """

    def __init__(self, rel_pos, mod_type, sequence, nterm=False, cterm=False):
        self.rel_pos = rel_pos
        self.mod_type = mod_type
        self.nterm = nterm
        self.cterm = cterm
        self.letter = sequence.pep_seq[rel_pos]
        self.abs_pos = [
            rel_pos + match.rel_pos
            for match in sequence.protein_matches
        ]
        self.exact = [
            match.exact
            for match in sequence.protein_matches
        ]

    def display_mod_type(self):
        """
        Return the mod_type in an abbreviated form (i.e. "p" for "Phospho")

        Returns
        -------
        str
        """
        if self.mod_type in ["Phospho"]:
            return "p"
        if self.mod_type in ["Carbamidomethyl"]:
            return "cm"

        return self.mod_type

    def __hash__(self):
        return hash(
            (
                self.rel_pos,
                self.mod_type,
                self.nterm,
                self.cterm,
                self.letter,
                tuple(self.abs_pos),
                tuple(self.exact),
            )
        )

    def __eq__(self, other):
        if not isinstance(other, Modification):
            raise TypeError()

        return (
            self.rel_pos,
            self.mod_type,
            self.nterm,
            self.cterm,
            self.letter,
            self.abs_pos,
            self.exact,
        ) == (
            other.rel_pos,
            other.mod_type,
            other.nterm,
            other.cterm,
            other.letter,
            other.abs_pos,
            other.exact,
        )


def allowed_mod_type(mod, any_letter=None, any_mod=None, letter_mod=None):
    """
    Check if a modification is of a type.

    Filters by letter, mod_type, or both.

    Parameters
    ----------
    mod : :class:`Modification<pyproteome.modification.Modification>`
    any_letter : set of str
    any_mod : set of str
    letter_mod : set of tuple of str, str
    """
    return (
        (
            any_letter is None or
            mod.letter.upper() in any_mod
        ) or (
            any_mod is None or
            mod.mod_type in any_letter
        ) or (
            letter_mod is None or
            (mod.letter.upper(), mod.mod_type) in letter_mod
        )
    )


def _extract_letter_mods(letter_mod_types=None):
    if letter_mod_types is None:
        return None, None, None

    any_letter = set()
    any_mod = set()
    letter_mod = set()

    for letter, mod_type in letter_mod_types:
        if letter is None and mod_type is None:
            raise Exception("Need at least one letter or mod type not None")
        elif letter is None and mod_type is not None:
            any_letter.add(mod_type)
        elif letter is not None and mod_type is None:
            any_mod.add(letter.upper())
        else:
            letter_mod.add((letter.upper(), mod_type))

    return any_letter, any_mod, letter_mod


def filter_mod_types(psms, letter_mod_types=None):
    """
    Filter a list of psms sequences with a given modification type.

    Parameters
    ----------
    psms : :class:`pandas.DataFrame`
    letter_mod_types : list of tuple of str or None, str or None, optional

    Returns
    -------
    :class:`pandas.DataFrame`
    """
    return psms[
        psms["Modifications"].apply(
            lambda x: bool(list(x.get_mods(letter_mod_types)))
        )
    ]



"""
This module provides functionality for comparing spectra with a list of
fragments.
"""

from . import masses


CID_TOL = 1000
HCD_TOL = 10

COLLISION_TOLS = {
    "CID": CID_TOL,
    "HCD": HCD_TOL,
}


class PeptideHit:
    """
    Attributes
    ----------
    mz : float
    intensity : float
    name : str or None
    score : int or None
    predicted_mz : float or None
    match_list : dict of str, tuple of (float, float)
    num_losses : int
    """
    def __init__(
        self, mz, intensity,
        name=None, score=None, predicted_mz=None, match_list=None,
    ):
        self.mz = mz
        self.intensity = intensity
        self.name = name
        self.score = score
        self.predicted_mz = predicted_mz
        self.match_list = match_list
        self.num_losses = name.count("-") if name else 0

    @property
    def frac_error(self):
        if self.predicted_mz is None:
            return 0

        return abs(self.predicted_mz - self.mz) / self.predicted_mz


def _calculate_ion_score(ion_name):
    score = 0

    if ion_name.startswith("MH"):
        score += 12
    elif ion_name.startswith("b_") or ion_name.startswith("y_"):
        score += 10

    score -= ion_name.count("-")

    return score


def compare_spectra(
    spectra, frag_ions, charge, c13_num,
    tol=None,
):
    """
    Parameters
    ----------
    spectra : :class:`pymzml.spec.Spectrum<spec.Spectrum>`
    frag_ions : dict of str, float
    charge : int
    c13_num : int
    tol : float, optional

    Returns
    -------
    list of :class:`PeptideHit<pycamv.compare.PeptideHit>`
        Peak assignments for each peak in spectra.
    """
    if tol is None:
        tol = CID_TOL

    max_y = max(spectra.i)

    # C13 isotope calculations
    # XXX: Support C13 isotopes
    delta_c13 = masses.exact_mass({"C": [-1, 1]})
    delta_iso = [delta_c13/i for i in range(1, charge + 1)]

    # Iterate over each peak, looking for fragments to assign to it
    out = []

    # Reprofiled Peaks? Centroided Peaks?
    for intensity, mz in spectra.centroidedPeaks:
        # intensity, mz = np.array(intensity)
        peak_candidates = {
            ion_name: (ion_mz, abs(ion_mz - mz))
            for ion_name, ion_mz in frag_ions.items()
            if abs(ion_mz - mz) / ion_mz < 1.5 * tol / 1e6
        }

        if not peak_candidates:
            # XXX: AA Identification?
            out.append(PeptideHit(mz, intensity))
            continue

        # Calculate ion scores based on their names
        scores = {
            ion_name: _calculate_ion_score(ion_name)
            for ion_name in peak_candidates
        }

        # XXX: 0.5? 1
        closest_ion = min(peak_candidates, key=lambda x: peak_candidates[x][1])
        scores[closest_ion] += 1

        best_ion = max(scores, key=lambda x: scores[x])

        retain = (
            intensity > 0.1 * max_y or
            best_ion.startswith("MH") or
            best_ion.startswith("b_") or
            best_ion.startswith("y_") or
            (
                best_ion.count("-") == 0 and
                intensity > 0.025 * max_y
            )
        )

        if not retain:
            best_ion = None

        out.append(
            PeptideHit(
                mz,
                intensity,
                name=best_ion,
                score=scores.get(best_ion, None),
                predicted_mz=peak_candidates.get(best_ion, None),
                match_list=peak_candidates,
            )
        )

    # Free up memory so we don't keep duplicate info about a spectra's peaks
    del spectra._peaks
    del spectra._mz
    del spectra._i
    del spectra._centroidedPeaks
    del spectra._reprofiledPeaks

    return out



"""
Provides functions for interacting with MS data through ProteoWizard.
"""

import logging
import os
import platform
import requests
import shutil
import subprocess
import sys
import tempfile

import pymzml

from pyproteome import paths


LOGGER = logging.getLogger("pycamv.proteowizard")


THIS_DIR = os.path.abspath(os.path.dirname(__file__))
PROTEOWIZARD_DIR = os.path.join(THIS_DIR, "ProteoWizard")

PROTEOWIZARD_VERSION = "3.0.9393"
PROTEOWIZARD_PATH = os.path.join(
    PROTEOWIZARD_DIR,
    "ProteoWizard {}".format(PROTEOWIZARD_VERSION),
)

PROTEOWIZARD_MSI_32_URL = (
    "https://www.dropbox.com/s/5q7flvb32vgdh7d/"
    "pwiz-setup-3.0.9393-x86.msi?dl=1"
)

PROTEOWIZARD_MSI_64_URL = (
    "https://www.dropbox.com/s/7bz3wpgjj7zd99a/"
    "pwiz-setup-3.0.9393-x86_64.msi?dl=1"
)


def fetch_proteowizard(url=None):
    if os.path.exists(PROTEOWIZARD_PATH):
        return

    LOGGER.info("ProteoWizard not installed, fetching now.")

    if platform.system() not in ["Windows"]:
        raise Exception("Proteowizard install not supported on your platform")

    if url is None:
        if platform.architecture()[0] == "64bit":
            url = PROTEOWIZARD_MSI_64_URL
        else:
            url = PROTEOWIZARD_MSI_32_URL

    tmpdir = tempfile.mkdtemp()
    out_path = os.path.join(tmpdir, url.rsplit("/", 1)[1].rsplit("?")[0])

    # Download the .msi file
    with open(out_path, mode="wb") as f:
        response = requests.get(url, stream=True)

        if not response.ok:
            raise Exception("Unable to download file: {}".format(response))

        for block in response.iter_content(1024):
            f.write(block)

    # Extract the msi file's contents
    extract_path = os.path.join(tmpdir, "msi_extract")
    cmd = [
        "msiexec",
        "/a",
        out_path,
        "/qb",
        "TARGETDIR=\"{}\"".format(extract_path),
    ]
    subprocess.check_call(" ".join(cmd), shell=True)

    # Copy the msi file's contents to PROTEOWIZARD_DIR
    src = os.path.join(
        extract_path,
        "PFiles",
        "ProteoWizard",
    )

    shutil.rmtree(PROTEOWIZARD_DIR)
    shutil.copytree(src, PROTEOWIZARD_DIR)
    shutil.rmtree(tmpdir)


def raw_to_mzml(basename, out_dir, scans=None, mz_window=None):
    """
    Covert a RAW file to .mzML using ProteoWizard.

    Parameters
    ----------
    basename : str
    out_dir : str
    scans : list of int, optional
    mz_window : list of int, optional

    Returns
    -------
    :class:`pymzml.run.Reader<run.Reader>`
    """
    fetch_proteowizard()

    ms_convert_path = os.path.join(PROTEOWIZARD_PATH, "msconvert.exe")
    raw_path = os.path.join(paths.MS_RAW_DIR, "{}.raw".format(basename))

    # Create a config file,
    config, config_path = tempfile.mkstemp(suffix=".txt", text=True)

    with os.fdopen(config, "w+") as config:
        if scans:
            config.write(
                "filter=\"scanNumber {}\"\n".format(
                    " ".join(str(scan) for scan in scans)
                )
            )

        if mz_window:
            config.write(
                "filter=\"mzWindow [{},{}]\"\n".format(
                    mz_window[0], mz_window[1],
                )
            )

    # Run msconvert to convert raw file to mzML
    LOGGER.info("Converting \"{}\" to .mzML format.".format(raw_path))

    cmd = [
        ms_convert_path,
        raw_path,
        "-o", out_dir,
        "--mzML",
        "-c", config_path,
    ]

    out = subprocess.check_output(cmd)
    LOGGER.debug(out.decode(sys.stdout.encoding))

    os.remove(config_path)

    # Read the file into memory using pymzml
    out_path = os.path.join(out_dir, "{}.mzML".format(basename))
    data = pymzml.run.Reader(
        out_path,
        extraAccessions=[
            ("MS:1000827", ["value"]),  # isolation window target m/z
            ("MS:1000828", ["value"]),  # isolation window lower offset
            ("MS:1000829", ["value"]),  # isolation window upper offset
            ("MS:1000512", ["value"]),  # filter string
        ],
    )

    return data




# Built-ins
from math import ceil

# IPython
from IPython.display import display

# Core data analysis libraries
from matplotlib import pyplot as plt
import numpy as np


def draw_logo(n_mers):
    """
    Draw a sequence logo.

    Parameters
    ----------
    n_mers : list of str
    """
    assert len(n_mers) >= 1

    n = len(n_mers[0])
    assert n % 2 == 1
    assert all(len(i) == n for i in n_mers)

    f, ax = plt.subplots()

    indices = np.arange(-n // 2, n // 2 + 1)
    bits = [{"A": 1, "S": 2}]

    ax.set_xlabel("Position")
    ax.set_xticks(indices)
    #ax.set_xticklabels

    ax.set_ylabel("Bits")
    max_bits = int(ceil(max(max(i.values()) for i in bits)))
    ax.set_yticks(np.arange(max_bits + 1))
    ax.set_ylim((0, max_bits + .5))

    from matplotlib.transforms import Affine2D

    for index, seq_bits in enumerate(bits):
        x_pos = indices[index]
        y_pos = 0

        for letter, size in seq_bits.items():
            tr = Affine2D().scale(1, 2)
            display(letter)

            plt.text(x_pos, y_pos, letter, transform=ax.transData)
            y_pos += size

    return f



"""
This module provides functionality for manipulating sequences.
"""

# Built-ins
import logging


LOGGER = logging.getLogger("pyproteome.sequence")


class ProteinMatch:
    """
    Contains information mapping a sequence onto a protein.

    Attributes
    ----------
    protein : :class:`Protein<pyproteome.protein.Protein>`
    rel_pos : int
    exact : bool
    """

    def __init__(self, protein, rel_pos, exact):
        self.protein = protein
        self.rel_pos = rel_pos
        self.exact = exact

    def __hash__(self):
        return hash(
            (
                self.protein,
                self.rel_pos,
                self.exact,
            )
        )

    def __eq__(self, other):
        if not isinstance(other, ProteinMatch):
            raise TypeError()

        return (
            self.protein,
            self.rel_pos,
            self.exact,
        ) == (
            other.protein,
            other.rel_pos,
            other.exact,
        )


class Sequence:
    """
    Contains information about a sequence and which proteins it matches to.

    Attributes
    ----------
    pep_seq : str
    protein_matches : list of :class:`ProteinMatch<pyproteome.sequence.ProteinMatch>`
    modifications : :class:`Modifications<pyproteome.modification.Modifications>`
    alt_hits : list of :class:`Sequence<pyproteome.sequence.Sequence>`
        This attribute indicates which peptides were identified as
        non-ambiguous subsequences of this peptide.
    """

    def __init__(self, pep_seq, protein_matches, modifications=None):
        """
        Parameters
        ----------
        pep_seq : str
        protein_matches : list of :class:`ProteinMatch<pyproteome.sequence.ProteinMatch>`
        modifications : :class:`Modifications<pyproteome.modification.Modifications>`, optional
        """
        self.pep_seq = pep_seq
        self.protein_matches = protein_matches
        self.modifications = modifications
        self.alt_hits = []

    def __hash__(self):
        return hash(
            (
                self.pep_seq,
                self.modifications,
            )
        )

    def __eq__(self, other):
        # In case of searching just by sequence
        if isinstance(other, str):
            return self._seq_with_modifications() == other

        if not isinstance(other, Sequence):
            raise TypeError()

        return (
            self.pep_seq,
            self.modifications,
        ) == (
            other.pep_seq,
            other.modifications,
        )

    def __contains__(self, other):
        if not isinstance(other, Sequence):
            raise TypeError(type(other))

        self_mods = list(self.modifications.skip_labels_iter())
        other_mods = list(other.modifications.skip_labels_iter())

        return (
            other.pep_seq.upper() in self.pep_seq.upper() and
            len(other.protein_matches) == len(self.protein_matches) and
            all(
                i.protein == j.protein
                for i, j in zip(other.protein_matches, self.protein_matches)
            ) and
            len(other_mods) == len(self_mods) and
            all(
                i.mod_type == j.mod_type and
                i.abs_pos == j.abs_pos and
                i.nterm == j.nterm and
                i.cterm == j.cterm
                for i, j in zip(
                    other_mods,
                    self_mods,
                )
            )
        )

    def _seq_with_modifications(self):
        string = self.pep_seq.upper()

        for mod in self.modifications.skip_labels_iter():
            string = string[:mod.rel_pos] + string[mod.rel_pos].lower() + \
                string[mod.rel_pos + 1:]

        return string

    def __str__(self):
        if not self.modifications:
            return self.pep_seq
        else:
            return self._seq_with_modifications()



"""
This module provides functionality for interacting with CAMV.

Currently limited to importing and outputing scan lists.
"""

# Built-ins
from __future__ import division

from collections import OrderedDict
import logging
from math import ceil
import os
import subprocess

# Core data analysis libraries
from IPython.display import display
import pandas as pd

from . import modification, paths, utils


LOGGER = logging.getLogger("pyproteome.camv")
THIS_DIR = os.path.dirname(os.path.abspath(__file__))
CAMV_PATH = utils.which("CAMV.exe")

if CAMV_PATH is None:
    CAMV_PATH = os.path.abspath(
        os.path.join(
            THIS_DIR, "..", "..",
            "CAMV", "CAMV", "for_redistribution_files_only", "CAMV.exe",
        )
    )


def load_camv_validation(basename):
    """
    Load validation data produced by CAMV.

    Parameters
    ----------
    basename : str

    Returns
    -------
    accepted : :class:`pandas.DataFrame`
    maybed : :class:`pandas.DataFrame`
    rejected : :class:`pandas.DataFrame`
    """
    accepted = None
    maybed = None
    rejected = None

    def _try_open_xls(path, existing=None):
        try:
            df = pd.read_csv(path, sep="\t")
        except OSError:
            return None
        else:
            LOGGER.info(
                "Loading CAMV validation data from \"{}\"".format(
                    os.path.join(*path.split(os.sep)[-2:]),
                )
            )
            in_name = os.path.splitext(os.path.basename(path))[0]
            df["Scan Paths"] = pd.Series(
                [set([in_name])] * len(df.index)
            )

            if existing is not None:
                df = pd.concat([existing, df])

            return df

    for filename in os.listdir(paths.CAMV_OUT_DIR):
        if filename.startswith(basename):
            base_dir = os.path.join(paths.CAMV_OUT_DIR, filename)

            accept_path = os.path.join(base_dir, "accept.xls")
            maybe_path = os.path.join(base_dir, "maybe.xls")
            reject_path = os.path.join(base_dir, "reject.xls")

            accepted = _try_open_xls(accept_path, existing=accepted)
            maybed = _try_open_xls(maybe_path, existing=maybed)
            rejected = _try_open_xls(reject_path, existing=rejected)

    return accepted, maybed, rejected


def output_scan_list(
    psms,
    basename="Scan List",
    letter_mod_types=None,
    scan_sets=None,
):
    """
    Write a list of scans to file.

    Can select for any peptides a given amino acid / modification type
    (i.e. letter_mod_types=[("Y", "Phospho")] ).

    Parameters
    ----------
    psms : :class:`pandas.DataFrame`
    basename : str, optional
    letter_mod_types : list of tuple of str or None, str or None, optional
    scan_sets : int, optional

    Returns
    -------
    :class:`pandas.DataFrame`
    dict of str, list of int
        Dictionary listing the file names and scans segmented into each file.
    """
    filtered = modification.filter_mod_types(
        psms,
        letter_mod_types=letter_mod_types,
    )

    scan_lists = OrderedDict()

    if len(filtered) == 0:
        return psms, scan_lists

    psms = psms.copy()
    psms["Scan Paths"] = pd.Series([set("")] * len(psms.index))

    scan_list = filtered.sort(
        columns=["Protein Group Accessions", "First Scan"],
        ascending=True,
    )[["First Scan"]]

    scan_list.drop_duplicates(
        subset="First Scan",
        inplace=True,
    )

    # Export as XLS
    utils.make_folder(paths.SCAN_LISTS_DIR)

    if scan_sets is None:
        scan_sets = int(ceil(len(scan_list) / 1000))
        LOGGER.info(
            "Splitting {} unique scans into {} lists".format(
                len(scan_list),
                scan_sets,
            )
        )

    assert scan_sets >= 0

    slice_sizes = int(ceil(len(scan_list) / scan_sets))

    for i in range(scan_sets):
        out_name = "{}-{}.xls".format(basename, i + 1)
        writer = pd.ExcelWriter(
            os.path.join(
                paths.SCAN_LISTS_DIR,
                out_name,
            )
        )
        lst = scan_list[i * slice_sizes:(i + 1) * slice_sizes]
        scan_lists[out_name] = lst["First Scan"].tolist()

        lst.to_excel(
            writer,
            index=False,
            header=False,
            sheet_name="Scan List",
        )
        writer.save()

        # XXX: Better way to do this?
        for index, row in psms[
            psms["First Scan"].isin(scan_lists[out_name])
        ].iterrows():
            psms.set_value(
                index,
                "Scan Paths",
                set([os.path.splitext(out_name)[0]]),
            )

    return psms, scan_lists


def _run_camv_get_file(
    raw_path, xml_path, output_dir,
    scan_path=None, save_path=None,
):
    """
    Run the CAMV "Get File" command and "Save Session" commands.

    Parameters
    ----------
    raw_path : str
    xml_path : str
    output_dir : str
    scan_path : str, optional
    save_path : str, optional

    Returns
    -------
    str
    """
    LOGGER.info(
        (
            "Running CAMV import on \"{}\", \"{}\", \"{}\","
            " scans=\"{}\", save=\"{}\""
        ).format(
            os.path.basename(raw_path),
            os.path.basename(xml_path),
            os.path.split(output_dir)[0],
            os.path.basename(scan_path) if scan_path else "",
            os.path.basename(save_path) if save_path else "",
        )
    )
    cmd = [
        CAMV_PATH,
        "import", "true",
        "raw_path", raw_path,
        "xml_path", xml_path,
        "out_path", output_dir,
    ]

    if scan_path:
        cmd += ["sl_path", scan_path]

    if save_path:
        cmd += [
            "save", "true",
            "session_path", save_path,
        ]

    cmd += ["exit", "true"]

    output = subprocess.check_output(cmd)

    return output


def _run_camv_export(save_path):
    """
    Run CAMV "Load Session" and "Export" commands.

    Parameters
    ----------
    save_path : str

    Returns
    -------
    str
    """
    LOGGER.info(
        "Running CAMV export on \"{}\""
        .format(
            save_path
        )
    )

    cmd = [
        CAMV_PATH,
        "load", "true",
        "session_path", save_path,
        "export", "true",
        "exit", "true",
    ]

    output = subprocess.check_output(cmd)

    return output


def run_camv_validation(scan_lists, force=False):
    """
    Run CAMV on a list of scans.

    Does the initial step of converting, importing, and processing scans.

    Parameters
    ----------
    scan_lists : dict of str, list of int
    force : bool, optional
    """
    for scan_path, scan_list in scan_lists.items():
        # Build a list of paths
        scan_path = os.path.join(
            paths.SCAN_LISTS_DIR, scan_path,
        )
        file_name = os.path.basename(scan_path)
        base_name = file_name.rsplit("-", 1)[0]
        raw_path = os.path.join(
            paths.MS_RAW_DIR, base_name + ".raw",
        )
        mascot_xml_path = os.path.join(
            paths.MASCOT_XML_DIR, base_name + ".xml",
        )
        save_path = os.path.join(
            paths.CAMV_SESS_DIR, os.path.splitext(file_name)[0] + ".mat"
        )

        if not force and \
           os.path.exists(save_path) and \
           os.stat(save_path).st_size >= 2 ** 12:
            continue

        # Run CAMV
        _run_camv_get_file(
            raw_path,
            mascot_xml_path,
            paths.CAMV_OUT_DIR,
            scan_path=scan_path,
            save_path=save_path,
        )

        # Check save files exists and is >= 4 MB
        if not os.path.exists(save_path) or \
           os.stat(save_path).st_size < 2 ** 12:
            raise Exception(
                "CAMV did not create a save file for {}".format(save_path)
            )


def run_camv_export(scan_lists=None):
    """
    Run CAMV export command.

    Creates a list of excel files for accept / maybe / reject peptides using
    any saved CAMV sessions. If scan_lists is None, everything in the "CAMV
    Sessions" directory will be exported.

    Parameters
    ----------
    scan_lists : dict of str, list of int, optional
    """
    if scan_lists is None:
        scan_paths = [
            path
            for path in os.listdir(paths.CAMV_SESS_DIR)
            if path.endswith(".mat")
        ]
    else:
        scan_paths = list(scan_lists.keys())

    for scan_path in scan_paths:
        file_name = os.path.basename(scan_path)
        save_path = os.path.join(
            paths.CAMV_SESS_DIR, os.path.splitext(file_name)[0] + ".mat"
        )
        display(_run_camv_export(save_path))



"""
This module provides functionality for finding motifs in sequences.

Functionality includes n-mer generation.
"""

from __future__ import division

# Built-ins
from collections import defaultdict
import logging
import os
import re

# Core data analysis libraries
try:
    import Bio.Alphabet.IUPAC
    import Bio.Seq
    import Bio.motifs
except ImportError:
    Bio = None
import pandas as pd
from scipy.stats import hypergeom

from . import sequence


LOGGER = logging.getLogger("pyproteome.motif")


def generate_n_mers(
    sequences, n=15,
    all_matches=True,
    fill_left="A",
    fill_right="A",
    letter_mod_types=None,
):
    """
    Generate n-mers around all sites of modification in sequences.

    Parameters
    ----------
    sequences : list of :class:`Sequence<pyproteome.sequence.Sequence>`
    n : int, optional
    all_matches : bool, optional
        Generate n-mers for all protein matches else just the first match.
    fill_left : str, optional
    fill_right : str, optional
    """
    # Check n is odd
    assert n % 2 == 1

    def _n_mer_from_sequence(full_seq, abs_pos):
        return (
            fill_left * (n // 2 - abs_pos) +
            full_seq[max([abs_pos - n // 2, 0]):abs_pos] +
            full_seq[abs_pos].lower() +
            full_seq[abs_pos + 1:abs_pos + n // 2 + 1] +
            fill_right * (abs_pos - len(full_seq) + n // 2 + 1)
        )

    return (
        _n_mer_from_sequence(
            match.protein.full_sequence,
            mod.rel_pos + match.rel_pos,
        )
        for seq in sequences
        for mod in seq.modifications.get_mods(letter_mod_types)
        for match in seq.protein_matches[:None if all_matches else 1]
    )


class Motif:
    """
    Contains a motif that may match to one or more protein sequences.

    Matches include the regular single-letter amino acid names as well as
    phosphosites for serine, threonine, and tyrosine, non-polar amino acids,
    and positively and negatively charged amino acids.

    Attributes
    ----------
    motif : str

    Examples
    --------
    >>> import pyproteome
    >>> motif = pyproteome.Motif("O..x.-+")
    >>> "IEFyFER" in motif
    True
    >>> "IEFyFED" in motif
    False
    >>> "FFFFFFR" in motif
    False
    """
    def __init__(self, motif):
        # Set _motif and _re explicitly, py2.7 does not call motif's setter
        # if we just set self.motif here...
        self._motif = motif
        self._re = self._compile_re(motif)

    char_mapping = {
        "x": "st",
        "O": "MILV",
        "-": "DE",
        "+": "RK",
        ".": "ACDEFGHIKLMNPQRSTVWYystO-+x",
    }

    @property
    def motif(self):
        return self._motif

    @motif.setter
    def motif(self, value):
        self._motif = value
        self._re = self._compile_re(value)

    def _compile_re(self, motif):
        ret = ""

        for char in motif:
            if char == "." or char not in self.char_mapping:
                ret += char
            else:
                ret += "[{}{}]".format(self.char_mapping[char], char)

        return re.compile(ret)

    def generate_children(self):
        for index, x in enumerate(self.motif):
            if x != ".":
                continue

            for char in self.char_mapping.get(x, ""):
                yield Motif(
                    self.motif[:index] + char + self.motif[index + 1:],
                )

    def generate_pairwise_children(self, hit_list):
        # First build a list of all of the AAs in the hits for this motif. We
        # don't want to generate child motifs that do not at least match any
        # peptides in that list.
        aa_is = dict(
            (
                (i, j),
                set((hit[i], hit[j]) for hit in hit_list[self]),
            )
            for i in range(len(self.motif))
            for j in range(i + 1, len(self.motif))
        )

        for key, val in aa_is.items():
            to_add = set()
            for special_char in "O-+x":
                for aa_i, aa_j in val:
                    if aa_i in self.char_mapping[special_char]:
                        to_add.add((special_char, aa_j))
                    if aa_j in self.char_mapping[special_char]:
                        to_add.add((aa_i, special_char))
                    if aa_i in self.char_mapping[special_char] and \
                       aa_j in self.char_mapping[special_char]:
                        to_add.add((special_char, special_char))
            val.update(to_add)

        # Then generate motifs with AAs that match the hit list
        for i, char_i in enumerate(self.motif):
            if char_i != ".":
                continue

            for j, char_j in enumerate(self.motif[i + 1:], start=i + 1):
                if char_j != ".":
                    continue

                for aa_i, aa_j in aa_is[(i, j)]:
                    yield Motif(
                        self.motif[:i] + aa_i +
                        self.motif[i + 1:j] + aa_j +
                        self.motif[j + 1:]
                    )

    def __repr__(self):
        return "<pyproteome.Motif: {}>".format(self.motif)

    def __str__(self):
        return self.motif

    def __hash__(self):
        return hash(self.motif)

    def __contains__(self, other):
        if isinstance(other, Motif):
            other = other.motif

        if not isinstance(other, str):
            raise TypeError

        if len(self.motif) != len(other):
            raise ValueError("Incompatible sequence lengths.")

        return self.match(other)

    def __eq__(self, other):
        if isinstance(other, Motif):
            return self.motif == other.motif

        if not isinstance(other, str):
            raise TypeError

        if len(self.motif) != len(other):
            raise ValueError("Incompatible sequence lengths.")

        return self.match(other)

    def __lt__(self, other):
        if not isinstance(other, Motif):
            raise TypeError

        return self.motif < other.motif

    def match(self, other):
        """
        Match a given sequence to this motif.

        Parameters
        ----------
        other : str

        Returns
        -------
        bool
        """
        return bool(self._re.match(other))


def _motif_sig(fore_hits, fore_size, back_hits, back_size):
    return (
        1 -
        hypergeom.cdf(fore_hits, back_size, back_hits, fore_size) +
        hypergeom.pmf(fore_hits, back_size, back_hits, fore_size)
    )


def motif_enrichment(
    foreground, background,
    sig_cutoff=0.01, min_fore_hits=0,
    start_letters=None, letter_mod_types=None,
    motif_length=15,
):
    """
    Calculate motifs significantly enriched in a list of sequences.

    Parameters
    ----------
    foreground : list of :class:`Sequence<pyproteome.sequence.Sequence>`
    background : list of :class:`Sequence<pyproteome.sequence.Sequence>`
    sig_cutoff : float, optional
    min_motifs : int, optional
    start_letters : list of str, optional
    letter_mod_types : list of tuple of str, str
    motif_length : int, optional

    Returns
    -------
    :class:`pandas.DataFrame`

    Notes
    -----
    .. [1] Joughin, Brian a et al. "An Integrated Comparative Phosphoproteomic
           and Bioinformatic Approach Reveals a Novel Class of MPM-2 Motifs
           Upregulated in EGFRvIII-Expressing Glioblastoma Cells." Molecular
           bioSystems 5.1 (2009): 59-67.
    """
    def _motif_stats(motif):
        fore_hits, back_hits = done[motif]

        # Re-calculate the P enrichment score associated with this motif
        p_value = _motif_sig(fore_hits, fore_size, back_hits, back_size)

        return (
            motif,
            fore_hits, fore_size,
            back_hits, back_size,
            p_value,
        )

    def _count_occurences(motif, parent, hit_list):
        hits = [
            i
            for i in hit_list[parent]
            if motif.match(i)
        ]
        hit_list[motif] = hits
        return len(hits)

    def _check_anchestry(motif, parent):
        for index, char in enumerate(motif.motif):
            if index == motif_length // 2 or char != ".":
                continue

            new_motif = Motif(
                motif.motif[:index] + "." + motif.motif[index + 1:]
            )

            if new_motif != parent and new_motif in done:
                return True

        return False

    def _is_a_submotif_with_same_size(submotif, fore_hits):
        """
        Test whether the 'submotif' is an ancestor of any known motif, and
        matches the same number of foreground sequences.
        """
        # We want to know if this "submotif" contains any completed motifs
        return any(
            checked in submotif
            for checked, checked_hits in done.items()
            if fore_hits == checked_hits[0] and checked != submotif
        )

    def _search_children(children, parent=None):
        """
        Run a depth-first search over a given motif.
        """
        motif_hits = set()

        for motif in children:
            if motif in visited:
                failed["checkpat"] += 1
                continue

            # Mark motif as visited
            visited.add(motif)

            if motif in done:
                failed["done"] += 1
                continue

            # Calculate the number of foreground hits
            fore_hits = _count_occurences(motif, parent, fg_hit_list)

            if fore_hits < min_fore_hits:
                failed["count"] = 1
                del fg_hit_list[motif]
                continue

            # Check if we've already completed the search for a parent of this
            # motif
            if parent and _check_anchestry(motif, parent):
                failed["anchestry"] += 1
                del fg_hit_list[motif]
                continue

            # Shortcut calculating back-hits if we can help it
            best_p = _motif_sig(fore_hits, fore_size, fore_hits, back_size)
            if best_p > sig_cutoff:
                failed["bestsig"] += 1
                del fg_hit_list[motif]
                continue

            if _is_a_submotif_with_same_size(motif, fore_hits):
                failed["sub"] += 1
                del fg_hit_list[motif]
                continue

            # Calculate the number of background hits
            back_hits = _count_occurences(motif, parent, bg_hit_list)

            # Check the signifiance of the motif
            p_value = _motif_sig(fore_hits, fore_size, back_hits, back_size)
            if p_value > sig_cutoff:
                failed["sig"] += 1
                del fg_hit_list[motif]
                del bg_hit_list[motif]
                continue

            # If we get this far, we have a hit!
            motif_hits.add(motif)
            failed["succeed"] += 1
            done[motif] = (fore_hits, back_hits)

            # Now try adding to the list of modifications on the motif to find
            # something more specific.
            motif_hits.update(
                _search_children(
                    children=motif.generate_children(),
                    parent=motif,
                )
            )
            motif_hits.update(
                _search_children(
                    children=motif.generate_pairwise_children(fg_hit_list),
                    parent=motif,
                )
            )

            del fg_hit_list[motif]
            del bg_hit_list[motif]

        return motif_hits

    def _filter_less_specific(prev_pass):
        return set(
            motif
            for motif in prev_pass
            if not any(
                # other in motif: other is +more+ specific than motif
                other in motif and done[other][0] >= done[motif][0]
                for other in prev_pass
                if other != motif
            )
        )

    def _make_nmers(lst):
        if isinstance(next(iter(lst)), sequence.Sequence):
            return list(
                generate_n_mers(
                    lst,
                    n=motif_length,
                    letter_mod_types=letter_mod_types,
                )
            )

        assert all(len(i) == motif_length for i in lst)
        return lst

    fore_size = len(foreground)
    back_size = len(background)

    assert fore_size > 0
    assert back_size > 0
    assert motif_length % 2 == 1

    if letter_mod_types is None:
        letter_mod_types = [(None, "Phospho")]

    foreground = _make_nmers(foreground)
    background = _make_nmers(background)

    LOGGER.info(
        "Starting analysis, n={}, N={}".format(
            fore_size, back_size,
        )
    )
    LOGGER.info(
        "Motif sig cutoff: {}, Minimum hits: {}".format(
            sig_cutoff, min_fore_hits,
        )
    )

    visited, done = set(), {}
    fg_hit_list = defaultdict(list)
    bg_hit_list = defaultdict(list)
    failed = defaultdict(int)

    # Set the starting motif and begin adding modifications to it.
    # Note: Order matters, put less specific to the end of this list
    if start_letters is None:
        start_letters = ["k", "y", "s", "t", "x"]

    starts = [
        Motif(
            "." * (motif_length // 2) + letter + "." * (motif_length // 2)
        )
        for letter in start_letters
    ]

    LOGGER.info(
        "Starting Motifs: {}".format(", ".join(str(i) for i in starts))
    )

    for start in starts:
        fg_hit_list[start] = foreground
        bg_hit_list[start] = background

    first_pass = set()

    try:
        for start in starts:
            first_pass.update(
                _search_children(
                    children=start.generate_children(),
                    parent=start,
                )
            )

            first_pass.update(
                _search_children(
                    children=start.generate_pairwise_children(fg_hit_list),
                    parent=start,
                )
            )

            del fg_hit_list[start]
            del bg_hit_list[start]
    except KeyboardInterrupt:
        return
    finally:
        # Output some debugging information to compare with Brian's output
        LOGGER.info(
            "\n" + "\n".join(
                "FAIL_{}: {}".format(key.upper(), failed[key])
                for key in [
                    "done", "sub", "count", "sig", "bestsig",
                    "anchestry", "checkpat",
                ]
            ) + "\nSUCCEED: {}".format(failed["succeed"])
        )

    LOGGER.info("First pass: {} motifs".format(len(first_pass)))

    second_pass = _filter_less_specific(first_pass)

    LOGGER.info(
        "Second pass (less specific motifs removed): {} motifs"
        .format(len(second_pass))
    )

    # Finally prepare the output as a sorted list with the motifs and their
    # associated statistics.
    df = pd.DataFrame(
        data=[_motif_stats(motif) for motif in second_pass],
        columns=[
            "Motif",
            "Foreground Hits",
            "Foreground Size",
            "Background Hits",
            "Background Size",
            "p-value",
        ],
    )
    df.sort_values(by=["p-value", "Motif"], inplace=True)
    df.reset_index(drop=True)

    return df


def motif_logo(
    data, letter_mod_types=None,
    folder_name=None, filename="Motif.svg",
):
    """
    Create a sequence logo figure.

    Logos are created based on the frequencies of peptides in a data set.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`
    letter_mod_types : list of tuple of str, str
    folder_name : str, optional
    filename : str, optional
    """
    if folder_name and filename:
        filename = os.path.join(folder_name, filename)

    alpha = Bio.Alphabet.IUPAC.protein
    m = Bio.motifs.create(
        [
            Bio.Seq.Seq(seq.upper(), alphabet=alpha)
            for seq in generate_n_mers(
                data.psms["Sequence"],
                letter_mod_types=letter_mod_types,
            )
        ],
        alphabet=alpha,
    )
    m.weblogo(filename, format="svg", logo_title="Testing")



"""
This module provides functionality for calculating the masses of peptide
fragments.
"""

from collections import defaultdict

import numpy as np

from . import masses, ms_labels


def _sequence_mass(pep_seq):
    return sum(
        masses.AMINO_ACIDS[letter] +
        sum(
            masses.MODIFICATIONS[letter, mod]
            for mod in mods
        )
        for letter, mods in pep_seq
    )


def _sequence_name(pep_seq):
    return "".join(
        letter
        for letter, mods in pep_seq
        if letter not in ["N-term", "C-term"]
    )


def internal_fragment_ions(pep_seq, aa_losses=None, mod_losses=None):
    """
    Calculate the m/z of all internal fragmenets of a peptide.

    Parameters
    ----------
    pep_seq : list of tuple of (str, list of str)
        A list of peptide letters and each residue's modification(s).
    aa_losses : list of str, optional
        Potential neutral losses for each fragment (i.e. Water, amine, CO).
        List is composed of neutral loss names.
    mod_losses : dict of tuple of (str, str), list of str
        Potential neutral losses for modified amino acids (i.e. pY-HPO_3).
        Dictionary should map (letter, modification) to a list of neutral
        loss names.

    Returns
    -------
    dict of str, float
        Dictionary mapping ion names to ion m/z's.
    """
    if aa_losses is None:
        aa_losses = [
            "H_2O",
            "NH_3",
            "CO",
        ]

    if mod_losses is None:
        mod_losses = {
            ("M", "Oxidation"): ["SOCH_4"],
            ("M", "Dioxidation"): ["SO_2CH_4"],
            ("S", "Phospho"): ["H_3PO_4"],
            ("T", "Phospho"): ["H_3PO_4"],
            ("Y", "Phospho"): ["HPO_3", "HPO_3-H_2O"],
        }

    pep_seq = [
        (letter, mods)
        for letter, mods in pep_seq
        if letter not in ["N-term", "C-term"]
    ]

    frag_masses = {}

    for start in range(2, len(pep_seq)):
        for end in range(start + 1, len(pep_seq)):
            # Only add the mass of an N-terminus, cleavage will be between
            # C=O and N-H bond, adding a hydrogen to N-H
            fragment = [("N-term", [])] + pep_seq[start:end]

            mass = _sequence_mass(fragment)
            name = _sequence_name(fragment)

            frag_masses[name] = mass

            for loss in aa_losses:
                frag_masses[name + loss] = mass - masses.MASSES[loss]

            # M, ST, Y losses
            for (letter, mod), losses in mod_losses.items():
                if not any(
                    letter == l and mod in mods
                    for l, mods in fragment
                ):
                    continue

                for loss in losses:
                    frag_masses[name + loss] = mass - masses.MASSES[loss]

    return frag_masses


def _get_frag_masses(pep_seq):
    return [
        _sequence_mass([pep_seq[index]])
        for index in range(len(pep_seq))
    ]


def _charged_m_zs(name, mass, max_charge):
    for charge in range(1, max_charge + 1):
        yield (
            (
                name.split("-")[0] +
                (
                    "^{{{:+}}}".format(charge)
                    if charge > 1 else
                    "^{+}"
                ) +
                "-".join([] + name.split("-")[1:])
            ),
            (mass + charge * masses.PROTON) / charge,
        )


def _b_y_ions(
    pep_seq, frag_masses, fragment_max_charge,
    any_losses, aa_losses, mod_losses,
):
    def _generate_losses(seq, losses=None, max_depth=2):
        if losses is None:
            losses = defaultdict(int)

        if max_depth < 1:
            yield losses
            return

        for loss in any_losses:
            new_losses = losses.copy()
            new_losses[loss] += 1

            yield new_losses

            child_losses = _generate_losses(
                seq, new_losses,
                max_depth=max_depth - 1,
            )

            for new_losses in child_losses:
                yield new_losses

        for aa, a_losses in aa_losses.items():
            for index, (letter, mods) in enumerate(seq):
                if aa != letter:
                    continue

                new_seq = seq[:index] + seq[index + 1:]

                for loss in a_losses:
                    new_losses = losses.copy()
                    new_losses[loss] += 1

                    yield new_losses

                    child_losses = _generate_losses(
                        new_seq, new_losses,
                        max_depth=max_depth - 1,
                    )

                    for new_losses in child_losses:
                        yield new_losses

        for (aa, mod), m_losses in mod_losses.items():
            for index, (letter, mods) in enumerate(seq):
                if aa != letter or mod not in mods:
                    continue

                new_seq = seq[:index] + seq[index + 1:]

                for loss in m_losses:
                    new_losses = losses.copy()
                    new_losses[loss] += 1

                    yield new_losses

                    child_losses = _generate_losses(
                        new_seq, new_losses,
                        max_depth=max_depth - 1,
                    )

                    for new_losses in child_losses:
                        yield new_losses

    def _generate_ions(seq, mass, basename):
        # b/y ions
        charged_mzs = _charged_m_zs(basename, mass, fragment_max_charge)

        for name, mz in charged_mzs:
            yield name, mz

        # b/y ions with losses
        for losses in _generate_losses(seq):
            loss_mass = sum(
                masses.MASSES[loss_name] * count
                for loss_name, count in losses.items()
            )
            loss_name = "-".join(
                "{}{}".format(
                    "{} ".format(count) if count > 1 else "",
                    loss_name,
                )
                for loss_name, count in losses.items()
            )

            charged_mzs = _charged_m_zs(
                "{}-{}".format(basename, loss_name),
                mass - loss_mass,
                fragment_max_charge,
            )

            for name, mz in charged_mzs:
                yield name, mz

    for index in range(2, len(pep_seq) - 1):
        # XXX: iTRAQ / TMT y-adducts?
        # y ion: 1 hydrogen added to NH group, one hydrogen on K/R
        base_ions = {
            "a_{{{}}}".format(index - 1):
                sum(frag_masses[:index]) - masses.MASSES["CO"],
            "b_{{{}}}".format(index - 1):
                sum(frag_masses[:index]),
            "y_{{{}}}".format(index - 1):
                sum(frag_masses[index:]) + 2 * masses.PROTON,
        }

        for ion_name, ion_mass in base_ions.items():
            for name, mz in _generate_ions(
                pep_seq[:index], ion_mass, ion_name,
            ):
                yield name, mz


def _label_ions(pep_seq):
    label_mods = [
        mod
        for mod in pep_seq[0][1]
        if mod in ms_labels.LABEL_NAMES
    ]

    for mod in label_mods:
        for name, mz in zip(
            ms_labels.LABEL_NAMES[mod],
            ms_labels.LABEL_MASSES[mod],
        ):
            yield name, mz


def _parent_ions(frag_masses, parent_max_charge):
    parent_mass = sum(frag_masses)

    for name, mz in _charged_m_zs("MH", parent_mass, parent_max_charge):
        yield name, mz


def _py_ions(pep_seq):
    ions = {}

    if any(
        letter == "Y" and "Phospho" in mods
        for letter, mods in pep_seq
    ):
        ions["pY"] = masses.IMMONIUM_IONS["Y"] + \
            masses.MODIFICATIONS["Y", "Phospho"]

    return ions


def fragment_ions(
    pep_seq, charge,
    parent_max_charge=None, fragment_max_charge=None,
    any_losses=None, aa_losses=None, mod_losses=None,
):
    """
    Calculate the m/z of all ions generated by a peptide.

    Parameters
    ----------
    pep_seq : str
    charge : int
    parent_max_charge : int, optional
    fragment_max_charge : int, optional
    any_losses : list of str, optional
    aa_losses : dict of str, str, optional
        Potential neutral losses for each fragment (i.e. Water, amine, CO).
        List is composed of neutral loss names.
    mod_losses : dict of tuple of (str, str), str, optional
        Potential neutral losses for modified amino acids (i.e. pY-HPO_3).
        Dictionary should map (letter, modification) to a list of neutral
        loss names.

    Returns
    -------
    dict of int, dict of str, float
        Dictionary mapping fragment position to a dictionary mapping ion names
        to ion m/z's.
    """
    assert "N-term" == pep_seq[0][0]
    assert "C-term" == pep_seq[-1][0]

    if parent_max_charge is None:
        parent_max_charge = charge

    if fragment_max_charge is None:
        # This correct?
        fragment_max_charge = parent_max_charge - 1

    if any_losses is None:
        any_losses = [
            "H_2O",
            "NH_3",
        ]

    if aa_losses is None:
        aa_losses = {}

    if mod_losses is None:
        mod_losses = {
            ("S", "Phospho"): ["H_3PO_4"],
            ("T", "Phospho"): ["H_3PO_4"],
            ("Y", "Phospho"): ["HPO_3", "HPO_3-H_2O"],
            ("M", "Oxidation"): ["SOCH_4"],
            ("M", "Dioxidation"): ["SO_2CH_4"],
        }

    # First calculate the masses of each residue along the backbone
    frag_masses = _get_frag_masses(pep_seq)

    frag_ions = {}

    # Get b/y (and associated a/c/x/z) ions
    frag_ions.update(
        _b_y_ions(
            pep_seq, frag_masses, fragment_max_charge,
            any_losses, aa_losses, mod_losses,
        )
    )

    # Get parent ions (i.e. MH^{+1})
    frag_ions.update(
        _parent_ions(frag_masses, parent_max_charge)
    )

    # Get TMT / iTRAQ labels
    frag_ions.update(
        _label_ions(pep_seq)
    )

    # Get pY peak
    frag_ions.update(
        _py_ions(pep_seq)
    )

    # Get internal fragments
    frag_ions.update(
        internal_fragment_ions(
            pep_seq,
            aa_losses=aa_losses,
            mod_losses=mod_losses,
        )
    )

    return frag_ions



"""
Provides functionality for interacting with MASCOT data.
"""

import os
import re
import xml.etree.ElementTree as ET

from scipy.misc import comb

from pyproteome import paths
from . import ms_labels


MASCOT_NS = {
    "mascot":
        "http://www.matrixscience.com/xmlns/schema/mascot_search_results_2",
}


class PeptideQuery:
    """
    Attributes
    ----------
    gi : str
    protein : str
    query : int
    pep_rank : int
    pep_score : float
    pep_exp_mz : float
    pep_exp_z : int
    pep_seq : str
    pep_var_mods : list of tuple of (int, str, list of str)
    scan : int
    num_comb : int
    """
    def __init__(
        self, gi, protein, query,
        pep_rank, pep_score, pep_exp_mz, pep_exp_z, pep_seq,
        pep_var_mods, scan,
    ):
        self.gi = gi
        self.protein = protein
        self.query = query
        self.pep_rank = pep_rank
        self.pep_score = pep_score
        self.pep_exp_mz = pep_exp_mz
        self.pep_exp_z = pep_exp_z
        self.pep_seq = pep_seq
        self.pep_var_mods = pep_var_mods
        self.scan = scan
        self.num_comb = self._calc_num_comb()

    def _unique_tuple(self):
        return (
            self.gi,
            self.query,
            self.pep_seq,
            self.scan,
        )

    def __hash__(self):
        return hash(self._unique_tuple())

    def __eq__(self, other):
        if not isinstance(other, PeptideQuery):
            raise TypeError(other)
        return self._unique_tuple() == other._unique_tuple()

    @property
    def get_label_mods(self):
        return [
            mod
            for count, mod, letters in self.pep_var_mods
            if mod in ms_labels.LABEL_NAMES and letters == ["N-term"]
        ]

    def _calc_num_comb(self):
        num_comb = 1

        for count, mod, letters in self.pep_var_mods:
            if mod == "Phospho" and letters == ["S", "T"]:
                letters = ["S", "T", "Y"]

            potential_mod_sites = sum(self.pep_seq.count(i) for i in letters)

            # Subtract sites that will be taken up by another modification
            # (i.e. Oxidation and Dioxidation of M)
            for o_count, o_mod, o_letters in self.pep_var_mods:
                if o_letters == letters and o_mod != mod:
                    potential_mod_sites -= o_count

            num_comb *= comb(potential_mod_sites, count)

        return num_comb


def _parse_letters(letters):
    """
    Turns a string of residue letters (i.e. "STY") into a list.

    Includes special provisions for N- and C-term modifications.

    Parameters
    ----------
    letters : str

    Returns
    -------
    list of str
    """
    if letters in ["N-term", "C-term"]:
        return [letters]

    return list(letters)


def _parse_mascot_2_4_1(root):
    fixed_mods = [
        i.text
        for i in root.findall(
            "mascot:fixed_mods/mascot:modification/mascot:name",
            MASCOT_NS,
        )
    ]
    variable_mods = [
        i.text
        for i in root.findall(
            "mascot:variable_mods/mascot:modification/mascot:name",
            MASCOT_NS,
        )
    ]

    scan_used = {}
    index = 0
    out = []

    for hit in root.findall("mascot:hits/mascot:hit", MASCOT_NS):
        accession = hit.find("mascot:protein", MASCOT_NS).get("accession")
        prot_desc = hit.find("mascot:protein/mascot:prot_desc", MASCOT_NS).text

        for peptide in hit.findall("mascot:protein/mascot:peptide", MASCOT_NS):
            query = int(peptide.get("query"))
            rank = int(peptide.get("rank"))
            pep_score = float(peptide.find("mascot:pep_score", MASCOT_NS).text)
            exp_mz = float(peptide.find("mascot:pep_exp_mz", MASCOT_NS).text)
            exp_z = int(peptide.find("mascot:pep_exp_z", MASCOT_NS).text)
            pep_seq = peptide.find("mascot:pep_seq", MASCOT_NS).text

            var_mods = peptide.find("mascot:pep_var_mod", MASCOT_NS).text

            if var_mods:
                # i.e. "2 Phospho (STY)""
                var_mods = [
                    re.match(
                        "((\d+) )?(.+) \((.+)\)",
                        mod.strip()
                    ).group(2, 3, 4)
                    for mod in var_mods.split(";")
                ]
                var_mods = [
                    (int(count) if count else 1, name, _parse_letters(letters))
                    for count, name, letters in var_mods
                ]
            else:
                var_mods = []

            scan = int(
                re.search(
                    "(scans:|Cmpd_)(\d+)",
                    peptide.find("mascot:pep_scan_title", MASCOT_NS).text
                ).group(2)
            )

            # scan_used
            if scan in scan_used:
                index_match, rank_match = scan_used[scan]

                if rank >= rank_match:
                    continue
                else:
                    del out[index_match]

            scan_used[scan] = (index, rank)
            out.append(
                PeptideQuery(
                    accession,
                    prot_desc,
                    query,
                    rank,
                    pep_score,
                    exp_mz,
                    exp_z,
                    pep_seq,
                    var_mods,
                    scan,
                )
            )
            index += 1

    out = sorted(out, key=lambda x: x.scan)

    return fixed_mods, variable_mods, out


def read_mascot_xml(xml_name):
    """
    Parse a MASCOT XML file.

    Parameters
    ----------
    xml_name : str
        Name of XML file in Mascot XMLs directory.

    Returns
    -------
    fixed_mods : list of str
    var_mods : list of str
    out : list of :class:`PeptideQuery<pycamv.mascot.PeptideQuery>`
    """
    xml_path = os.path.join(paths.MASCOT_XML_DIR, xml_name)
    tree = ET.parse(xml_path)
    root = tree.getroot()

    ms_version = root.find("mascot:header/mascot:MascotVer", MASCOT_NS).text
    ms_version = tuple(int(i) for i in ms_version.split("."))

    if ms_version >= (2, 4, 1):
        return _parse_mascot_2_4_1(root)
#     elif ms_version == "2.4.0":
#         return _parse_mascot_2_4_0(root)
#     elif ms_version == "2.3.02":
#         return _parse_mascot_2_3_02(root)
#     elif ms_version == "2.1.03":
#         return _parse_mascot_2_1_03(root)
    else:
        raise Exception("Unsupported Mascot Version: {}".format(ms_version))



"""
This module provides functionality for loading data sets.

Functionality includes loading CAMV and MASCOT / Discoverer data sets.
"""
# Built-ins
import logging
import os
import re

# Core data analysis libraries
import numpy as np
import pandas as pd

from . import camv, fetch_data, protein, sequence, modification, paths, utils


LOGGER = logging.getLogger("pyproteome.loading")
RE_PROTEIN = re.compile("([A-Za-z0-9\(\)\[\]\\/\',\. \-\+]+) OS=")
RE_MODIFICATION = re.compile("(N-Term|C-Term|([A-Z])([0-9]*))\((.*)\)")


def _filter_unassigned_rows(psms):
    """
    Remove rows from psms with unassigned peptides / proteins.

    Parameters
    ----------
    psms : :class:`pandas.DataFrame`

    Returns
    -------
    :class:`pandas.DataFrame`
    """
    return psms.dropna(
        axis=0,
        how="any",
        subset=[
            "Protein Descriptions",
            "Protein Group Accessions",
            "Sequence"
        ],
    )


def _extract_protein(prot_string):
    match = fetch_data.RE_ACCESSION.search(prot_string)
    if not match:
        raise Exception(
            "Unable to find accession in \"{}\"".format(prot_string)
        )
    return protein.Protein(accession=match.group(0))


def _extract_proteins_from_accessions(prots_string):
    """
    Extract a list of proteins from a string containing protein accessions.

    Parameters
    ----------
    prots_string : str

    Returns
    -------
    :class:`Proteins<pyproteome.protein.Proteins`
    """
    return protein.Proteins(
        proteins=[
            protein.Protein(
                accession=accession.strip(),
            )
            for accession in prots_string.split(";")
        ],
    )


def _extract_proteins_from_description(prots_string):
    """
    Extract a list of proteins from a string containing protein descriptions.

    Parameters
    ----------
    prots_string : str

    Returns
    -------
    :class:`Proteins<pyproteome.protein.Proteins`
    """
    return protein.Proteins(
        proteins=[
            protein.Protein(
                accession=accession,
            )
            for accession in fetch_data.RE_ACCESSION.findall(prots_string)
        ],
    )


def _extract_sequence(proteins, sequence_string):
    """
    Extract a Sequence object from a list of proteins and sequence string.

    Does not set the Sequence.modifications attribute.

    Parameters
    ----------
    proteins : list of :class:`Protein<pyproteome.protein.Protein>`
    sequence_string : str

    Returns
    -------
    list of :class:`Sequence<pyproteome.sequence.Sequence>`
    """
    prot_matches = []

    def _get_rel_pos(protein, pep_seq):
        seq = protein.full_sequence
        pep_pos = seq.find(pep_seq)
        exact = True

        if pep_pos < 0:
            pep_pos = utils.fuzzy_find(pep_seq, seq)
            exact = False

        return pep_pos, exact

    for prot in proteins:
        rel_pos, exact = _get_rel_pos(prot, sequence_string.upper())
        prot_matches.append(
            sequence.ProteinMatch(
                protein=prot,
                rel_pos=rel_pos,
                exact=exact,
            )
        )

    return sequence.Sequence(
        pep_seq=sequence_string,
        protein_matches=prot_matches,
    )


def _extract_modification(seq, mod_string):
    """
    Extract a single modification from mod_string.

    Parameters
    ----------
    seq : :class:`Sequence<pyproteome.sequence.Sequence>`
    mod_string : str

    Returns
    -------
    :class:`Modification<pyproteome.modification.Modification>`
    """
    def _get_pos(match):
        if match.group(1) == "N-Term":
            return 0, True, False
        elif match.group(1) == "C-Term":
            return len(seq.pep_seq) - 1, False, True
        else:
            return int(match.group(3)) - 1, False, False

    match = RE_MODIFICATION.match(mod_string)
    pos, nterm, cterm = _get_pos(match)
    letter, mod = match.group(2), match.group(4)

    if not nterm and not cterm:
        assert seq.pep_seq[pos].upper() == letter

    return modification.Modification(
        rel_pos=pos,
        mod_type=mod,
        nterm=nterm,
        cterm=cterm,
        sequence=seq,
    )


def _extract_modifications(sequence, mods_string):
    """
    Extract a structured list of modifications from mod_string.

    Parameters
    ----------
    sequence : :class:`Sequence<pyproteome.sequence.Sequence>`
    mod_string : str

    Returns
    -------
    :class:`Modifications<pyproteome.modification.Modifications>`
    """
    if isinstance(mods_string, float):
        mods_string = ""

    return modification.Modifications(
        mods=[
            _extract_modification(sequence, i.strip())
            for i in mods_string.split(";")
            if i
        ],
    )


def load_mascot_psms(basename, camv_slices=None):
    """
    Load a list of sequences from a file produced by MASCOT / Discoverer.

    Parameters
    ----------
    basenme : str
    camv_slices : int, optional

    Returns
    -------
    psms : :class:`pandas.DataFrame`
    scan_lists : dict of str, list of int
    filter_camv : bool
    """
    psms_path = os.path.join(
        paths.MS_SEARCHED_DIR,
        basename + "_psms.txt",
    )

    LOGGER.info(
        "Loading MASCOT peptides from \"{}\"".format(
            os.path.basename(psms_path),
        )
    )

    psms = pd.read_table(psms_path)
    psms = _filter_unassigned_rows(psms)

    # Pre-fetch UniProt data to speed up later queries
    fetch_data.prefetch_accessions(psms)

    psms["Proteins"] = pd.Series(
        [
            _extract_proteins_from_accessions(
                row["Protein Group Accessions"]
            )
            for index, row in psms.iterrows()
        ],
        index=psms.index,
    )
    psms["Sequence"] = pd.Series(
        [
            _extract_sequence(row["Proteins"], row["Sequence"])
            for index, row in psms.iterrows()
        ],
        index=psms.index,
    )
    psms["Modifications"] = pd.Series(
        [
            _extract_modifications(row["Sequence"], row["Modifications"])
            for index, row in psms.iterrows()
        ],
        index=psms.index,
    )

    psms.reset_index(inplace=True, drop=True)

    # Finally close the reference loop between sequences and modifications
    for index, row in psms.iterrows():
        row["Sequence"].modifications = row["Modifications"]

    # Output the phosphorylation scan list for CAMV
    psms, scan_lists = camv.output_scan_list(
        psms,
        basename=basename,
        letter_mod_types=[(None, "Phospho")],
        scan_sets=camv_slices,
    )

    psms["Validated"] = False

    # The load CAMV data to clear unwanted hits if available.
    accepted, maybed, rejected = camv.load_camv_validation(basename)

    if rejected is not None:
        # Remove any peptides that match the scan number and sequence
        # in the rejected list.
        reject_mask = np.zeros(psms.shape[0], dtype=bool)

        for index, row in psms.iterrows():
            # Check if this specific sequence and scan was rejected
            hit = np.logical_and(
                # Assuming First Scan always == Last Scan
                rejected["Scan"] == row["First Scan"],
                rejected["Sequence"] == row["Sequence"],
            )

            if hit.any():
                reject_mask[index] = True
                continue

            # Check if this scan was rejected and no sequences were accepted
            hit = (rejected["Scan"] == row["First Scan"]).any()
            if not hit:
                continue

            if accepted is not None:
                if np.logical_and(
                    accepted["Scan"] == row["First Scan"],
                    accepted["Sequence"] == row["Sequence"],
                ).any():
                    psms.loc[index, "Validated"] = True
                    continue

            if maybed is not None:
                if np.logical_and(
                    maybed["Scan"] == row["First Scan"],
                    maybed["Sequence"] == row["Sequence"],
                ).any():
                    continue

            reject_mask[index] = True

        psms = psms[~reject_mask].reset_index(drop=True)

    if accepted is not None:
        reject_mask = np.zeros(psms.shape[0], dtype=bool)

        for index, row in psms.iterrows():
            # Reject hits where the scan number is the same but the sequence
            # is different.
            hit = np.logical_and(
                accepted["Scan"] == row["First Scan"],
                accepted["Sequence"] != row["Sequence"],
            )
            if hit.any():
                reject_mask[index] = True

        psms = psms[~reject_mask].reset_index(drop=True)

    return psms, scan_lists


def load_validated_psms(filename):
    """
    Wrap load_camv_validation, returning only accepted peptides.

    Parameters
    ----------
    filename : str

    Returns
    -------
    :class:`pandas.DataFrame`
    """
    accepted, maybed, rejected = camv.load_camv_validation(filename)

    accepted["Validated"] = True
    return accepted



# -*- coding: utf-8 -*-
"""
This module provides functionality for interpreting BCA assays.
"""

# Built-ins
from collections import defaultdict
import logging
import os
import re

# Core data analysis libraries
from matplotlib import pyplot as plt
import matplotlib.cm as cm
import numpy as np
import pandas as pd
from scipy.stats import linregress

from . import paths

LOGGER = logging.getLogger("pyproteome.bca")
RE_ROW_COL = re.compile("([A-Z]+)(\d+)")


def _interpret_pos(pos):
    """
    Turn a position (i.e. "A1:B2") into integers for the row / column range.

    Parameters
    ----------
    pos : str

    Returns
    -------
    col_start : int
    col_end : int
    row_start : int
    row_end : int
    """
    lst = pos.split(":")

    m = RE_ROW_COL.match(lst[0])
    row_start, col_start = m.groups()
    col_start = int(col_start)

    if len(lst) > 1:
        m = RE_ROW_COL.match(lst[1])
        row_end, col_end = m.groups()
        col_end = int(col_end)
    else:
        row_end, col_end = None, None

    return col_start, col_end, row_start, row_end


def _next_chr(char):
    return chr(ord(char) + 1)


def interpret_bca_assay(
    xls_path,
    samples,
    volumes,
    standards=None,
    n=3,
    data_row_start=27,
):
    u"""
    Load and interpret a plate reader output.

    Fits a curve to the standards, checks data sanity, and calculates sample
    concentrations.

    Parameters
    ----------
    xls_path : str
        Path to output of Tecan reader. File should be located in
        paths.BCA_ASSAY_DIR.
    samples : list of tuple of (str, float, str)
        Sample names, dilutions, and positions within the plate. Positions
        should be supplied as either the left index of the first of three
        aliquots (i.e. "A4") or a range indicating the starting and ending row
        and column (i.e. "A4:B6").
    volumes : dict of str, float
        Volumes of samples, in μL.
    standards : list of (tuple of int, float), optional
        Locations of standards in table. Defaults to the first n (below)
        columns, rows A-G, with concentrations 1 - 0.0625 mg / mL, followed by
        two blanks of PBS.
    n : int, optional
        Number of replicates to use for standards and samples (Defaults to 3).
    data_row_start : int, optional
        Starting row in excel file for absorbance table.

    Returns
    -------
    total_protein : dict of str, tuple of (float, float)
        Sample total protein contents. Supplies the mean and standard deviation
        for the calculated protein content (in μg)
    concentrations : dict of list of float
        Individual concentrations calculations.

    Raises
    ------

    Examples
    --------
    >>> from pyproteome import bca
    >>> total_protein, _ = bca.interpret_bca_assay(
    ...     "bsa-sample.xlsx",
    ...     samples=[
    ...         ("3157 Hippocampus", 10, "A4:B6"),
    ...         ("3157 Cortex", 10, "C4:D6"),
    ...         ("3146 Cerebellum", 10, "E4:F6"),
    ...         ("3146 Hippocampus", 10, "G4:H6"),
    ...     ],
    ...     volumes={
    ...         "3157 Hippocampus": 3000,
    ...         "3157 Cortex": 3000,
    ...         "3146 Cerebellum": 3000,
    ...         "3146 Hippocampus": 3000,
    ...     },
    ... )
    >>> cortex = total_protein["3157 Cortex"]
    >>> print("{:.2f} +/- {:.2f} ug".format(cortex[0], cortex[1])
    8094.51 +/- 340.66 ug
    """
    xls_path = os.path.join(paths.BCA_ASSAY_DIR, xls_path)

    xls = pd.read_excel(
        xls_path,
        skiprows=(
            list(range(data_row_start)) +
            list(range(data_row_start + 9, data_row_start + 13))
        ),
        index_col=0,
    )

    if isinstance(xls, dict):
        assert len(xls) == 1
        xls = list(xls.values())[0]

    # Drop columns with NaN values
    xls = xls.dropna(
        axis=1,
        how="any",
    )

    std_start_col = xls.columns[0]

    if standards is None:
        standards = [
            ("A{}".format(std_start_col), 1.0),
            ("B{}".format(std_start_col), 0.5),
            ("C{}".format(std_start_col), 0.25),
            ("D{}".format(std_start_col), 0.125),
            ("E{}".format(std_start_col), 0.0625),
            ("F{}".format(std_start_col), 0),
            ("G{}".format(std_start_col), 0),
        ]

    # Extract out standards absorbances
    std_x = []
    std_y = []

    for pos, x in standards:
        col_start, col_end, row_start, row_end = _interpret_pos(pos)

        if col_end is None:
            col_end = col_start + 3

        if row_end is None:
            row_end = row_start

        row = row_start

        while row <= row_end:
            for col in range(std_start_col, std_start_col + 3):
                std_x.append(x)
                std_y.append(xls.ix[row][col])

            row = _next_chr(row)

    # Fit a linear regression model to standards
    slope, intercept, r, _, _ = linregress(std_x, std_y)

    # Plot these points on a graph
    f, ax = plt.subplots()
    ax.scatter(
        x=std_x,
        y=std_y,
        s=50,
        label="BSA Standards",
    )
    ax.plot(
        [0, max(std_x)],
        [intercept, intercept + slope * max(std_x)],
    )

    ax.set_title("BSA Standards $R^2$ = {:.3f}".format(r ** 2))
    ax.set_xlabel("Concentration (mg / mL)")
    ax.set_ylabel("Absorbance (AU)")

    # Sanity check standards
    if r ** 2 < 0.95:
        ax.legend(loc="lower right")
        raise Exception("R^2 of standards = {:.2f} (< 0.95)".format(r ** 2))

    # Extract out sample absorbances and concentrations
    absorbances = defaultdict(list)
    concentrations = defaultdict(list)
    total_protein = {}

    for name, dilution, pos in samples:
        col_start, col_end, row_start, row_end = _interpret_pos(pos)

        if col_end is None:
            col_end = col_start + 3

        if row_end is None:
            row_end = row_start

        row = row_start

        while row <= row_end:
            for col in range(col_start, col_end + 1):
                absorbance = xls.ix[row][col]
                concentration = (absorbance - intercept) / slope
                concentrations[name].append(concentration * dilution)
                absorbances[(name, dilution)].append(absorbance)

            row = _next_chr(row)

    # Plot samples and show to user
    colors = cm.rainbow(np.linspace(0, 1, len(absorbances)))
    for color, kv in zip(colors, absorbances.items()):
        key, val = kv
        name, dilution = key
        y = np.array(val)
        x = (y - intercept) / slope
        ax.scatter(x, y, s=100, label=name, color=color, marker="*")

    ax.legend(loc="lower right")

    # Sanity check concentrations

    # Calculate total sample protein content
    for name, volume in volumes.items():
        protein = np.array(concentrations[name]) * volume
        total_protein[name] = (
            protein.mean(),
            protein.std(),
        )

    return total_protein, concentrations



# Here is the function for calculating gene set enrichment scores.
# It calculates the association of each gene with a given phenotype
# and then generates the ES(S) scores for a given gene set.

def enrichment_scores(gene_changes, gene_set, p=1):
    assert len(gene_changes) >= len(gene_set)

    ranked_genes = sorted(gene_changes.items(), key=lambda x: x[1], reverse=True)

    n = len(gene_changes)
    n_h = len(gene_set)
    n_r = sum(
        abs(gene_changes.get(gene, 0)) ** p 
        for gene in gene_set
    )

    scores = [0] + [
        (abs(val) ** p) / n_r
        if gene in gene_set
        else
        - 1 / (n - n_h)
        for gene, val in ranked_genes
    ]

    return np.cumsum(scores)

# Here's a helper function to generate the plots seen below
def plot_enrichment(sorted_set, gene_sets, p=1, cols=1):
    f, axes = plt.subplots(len(gene_sets) // cols, cols, squeeze=False, sharey=True)
    f.set_size_inches(cols * 3.5, 4)
    axes = [i for j in axes for i in j]
    
    for index, ax, gene_set in zip(range(len(axes)), axes, gene_sets):
        ess = enrichment_scores(sorted_set, gene_set, p=p)
        ax.plot(ess)
        ax.set_title("{}, p={}\n(max: {:.2f}, min: {:.2f})"
                     .format("gene_set", p, max(ess), min(ess)))
        ax.axhline(0)
        ax.set_xlabel("Gene List Rank")
        
        if index % cols == 0:
            ax.set_ylabel("ES(S)")

def scores(sorted_set, gene_set, permute_n=1000):
    seed(0)
    es_s = max(
        enrichment_scores(
            sorted_set, 
            gene_set,
            p=1,
        ),
        key=lambda x: abs(x),
    )
    es_s_pi = np.array([
        max(
            enrichment_scores(
                dict(zip(np.random.permutation(list(sorted_set.keys())), sorted_set.values())),
                gene_set,
                p=1,
            ),
            key = lambda x: abs(x),
        )
        for i in range(permute_n)
    ])

    # Only select the subset of the binomial distribution that we care
    # about. Not 100% sure if this is kosher, statistically.
    if es_s >= 0:
        es_s_pi = es_s_pi[es_s_pi >= 0]
    else:
        es_s_pi = es_s_pi[es_s_pi <= 0]

    nes_s = es_s / np.mean(es_s_pi)
    nes_s_pi = es_s_pi / np.mean(es_s_pi)

    return nes_s, nes_s_pi

def plot_permutations(sorted_set, gene_sets, cols=1):
    f, axes = plt.subplots(len(gene_sets) // cols, cols, squeeze=False, sharey=True)
    f.set_size_inches(cols * 3.5, 4)
    axes = [i for j in axes for i in j]

    # Compute the distributions of all (S, \pi) and (S) for
    # later use in calculating the FDR q-value
    fdr_nes_s, fdr_nes_s_pi = array([]), array([])

    for gene_set in gene_sets:
        nes_s, nes_s_pi = scores(sorted_set, gene_set)
        fdr_nes_s = append(fdr_nes_s, [nes_s])
        fdr_nes_s_pi = append(fdr_nes_s_pi, nes_s_pi)

    # Plot each distribution and compute the true p and q-values
    for index, ax, gene_set in zip(range(len(axes)), axes, gene_sets):
        nes_s, nes_s_pi = scores(sorted_set, gene_set)
        p_value = sum(nes_s_pi >= nes_s) / nes_s_pi.size

        q_nominator = sum(fdr_nes_s_pi >= nes_s) / fdr_nes_s_pi.size
        q_denomator = sum(fdr_nes_s >= nes_s) / fdr_nes_s.size

        q_value = q_nominator / q_denomator

        ax.hist(nes_s_pi, normed=1)
        ax.axvline(nes_s, color="red")
        ax.set_title(
            (
                "Permuted phenotype labels\nGene set: {}\n"
                "p-value={:.3f}, FDR q-value={:.3f}"
            ).format("gene_set", p_value, q_value)
        )
        ax.set_xlabel("NES(S, $\pi$)")
        
        if index % cols == 0:
            ax.set_ylabel("Frequency")

def load_gene_set(path):
    with open(path) as f:
        return [
            line.split("#")[0].strip().upper()
            for line in f
        ]




"""
This module provides functionality for manipulating proteomics data sets.

Functionality includes merging data sets and interfacing with attributes in a
structured format.
"""

# Built-ins
from collections import OrderedDict
import copy
import logging

# Core data analysis libraries
import pandas as pd
import numpy as np
from scipy.stats import ttest_ind

from . import loading, utils


LOGGER = logging.getLogger("pyproteome.data_sets")


class DataSet:
    """
    Class that encompasses a proteomics data set.

    Includes peptide list, scan list, channels, groups, quantitative
    phenotypes, etc.

    Attributes
    ----------
    psms : :class:`pandas.DataFrame`
        Contains at least "Proteins", "Sequence", and "Modifications" columns.
    channels : dict of str, str
        Maps label channel to sample name.
    groups : dict of str, list of str
        Maps groups to list of sample names. The primary group is considered
        as the first in this sequence.
    phenotypes : dict of str, (dict of str, float)
        Primary key is the phenotype and its value is a dictionary mapping
        sample names to that phenotype's value.
    name : str
    enrichments : list of str
    tissues : list of str
    sets : int
        Number of sets merged into this data set.
    source : str
    scan_list : dict of str, list of int
    """
    def __init__(
        self, channels,
        psms=None,
        mascot_name=None,
        camv_name=None,
        groups=None, phenotypes=None,
        name="", enrichments=None, tissues=None,
        dropna=True,
        camv_slices=None,
        merge_duplicates=True,
        merge_subsets=False,
    ):
        """
        Initializes a data set.

        Parameters
        ----------
        channels : dict of str, str
        psms : :class:`pandas.DataFrame`, optional
        mascot_name : str, optional
        camv_name : str, optional
        groups : dict of str, list of str, optional
        phenotypes : dict of str, (dict of str, float), optional
        name : str, optional
        enrichments : list of str, optional
        tissues : list of str, optional
        dropna : bool, optional
        camv_slices : int, optional
        merge_duplicates : bool, optional
        merge_subsets : bool, optional
        """
        assert psms is not None or \
            mascot_name is not None or \
            camv_name is not None

        self.source = "unknown"
        self.scan_lists = None
        self.validated = False

        if mascot_name:
            psms, self.scan_lists = loading.load_mascot_psms(
                mascot_name,
                camv_slices=camv_slices,
            )
            self.source = "MASCOT"
        elif camv_name:
            psms = loading.load_validated_psms(
                camv_name,
            )
            self.source = "CAMV"

        self.psms = psms
        self.channels = channels
        self.groups = groups
        self.phenotypes = phenotypes
        self.name = name
        self.enrichments = enrichments
        self.tissues = tissues

        self.normalized = False
        self.sets = 1

        if dropna:
            LOGGER.info("Dropping channels with NaN values.")
            self.dropna(inplace=True)

        if merge_duplicates:
            LOGGER.info("Merging duplicate peptide hits together.")
            self._merge_duplicates()

        if merge_subsets:
            LOGGER.info("Merging peptide hits that are subsets together.")
            self._merge_subsequences()

    def copy(self):
        """
        Make a copy of self.

        Returns
        -------
        :class:`DataSet<pyproteome.data_sets.DataSet>`
        """
        new = copy.copy(self)
        new.psms = new.psms.copy()
        return new

    @property
    def enrichment(self):
        """
        The str version of self.enrichments

        Returns
        -------
        str
        """
        return "+".join(self.enrichments)

    @property
    def tissue(self):
        """
        The str version of self.tissues

        Returns
        -------
        str
        """
        return "+".join(self.tissues)

    def __str__(self):
        return (
            "<pyproteome.DataSet object" +
            (
                ": " + self.name
                if self.name else
                " at " + hex(id(self))
            ) +
            ">"
        )

    def __getitem__(self, key):
        if isinstance(key, slice):
            new = self.copy()
            new.psms = new.psms[key]
            return new

        if any(
            isinstance(key, i)
            for i in [str, list, tuple, pd.Series, np.ndarray]
        ):
            return self.psms[key]

        raise TypeError(type(key))

    def _merge_duplicates(self):
        channels = list(self.channels.keys())
        agg_dict = dict((channel, sum) for channel in channels)

        agg_dict["Validated"] = all
        agg_dict["Scan Paths"] = lambda x: set(
            path
            for paths in x
            for path in paths
        )

        self.psms = self.psms.groupby(
            by=[
                "Proteins",
                "Sequence",
                "Modifications",
            ],
            sort=False,
            as_index=False,
        ).agg(agg_dict)

    def _merge_subsequences(self):
        """
        Merges petides that are a subsequence of another peptide.

        Only merges peptides that contain the same set of modifications and
        that map to the same protein(s).
        """
        psms = self.psms

        # Find all proteins that have more than one peptide mapping to them
        for index, row in psms[
            psms.duplicated(subset="Proteins", keep=False)
        ].iterrows():
            seq = row["Sequence"]

            # Then iterate over each peptide and find other non-identical
            # peptides that map to the same protein
            for o_index, o_row in psms[
                np.logical_and(
                    psms["Proteins"] == row["Proteins"],
                    psms["Sequence"] != seq,
                )
            ].iterrows():
                # If that other peptide is a subset of this peptide, rename it
                if o_row["Sequence"] in seq:
                    psms.set_value(o_index, "Sequence", seq)
                    psms.set_value(o_index, "Modifications", seq.modifications)

        # And finally group together peptides that were renamed
        self._merge_duplicates()

    def __add__(self, other):
        """
        Concatenate two data sets.

        Combines two data sets, adding together the channel values for any
        common data.

        Parameters
        ----------
        other : :class:`DataSet<pyproteome.data_sets.DataSet>`

        Returns
        -------
        :class:`DataSet<pyproteome.data_sets.DataSet>`
        """
        return merge_data([self, other])

    def normalize(self, levels, inplace=False):
        """
        Normalizes a channel to given levels.

        Divides all channel values by a given level.

        Parameters
        ----------
        levels : dict of str, float
            Mapping of channel names to normalized levels.
        inplace : bool, optional
            Modify this data set in place.

        Returns
        -------
        :class:`DataSet<pyproteome.data_sets.DataSet>`
        """
        new = self

        # Don't normalize a data set twice!
        assert not new.normalized

        if not inplace:
            new = new.copy()

        new_channels = utils.norm(self.channels)

        for key, norm_key in zip(self.channels, new_channels):
            new.psms[norm_key] = new.psms[key] / levels[key]

        new.normalized = True
        new.channels = new_channels
        new.groups = OrderedDict(
            (key, utils.norm(val))
            for key, val in self.groups.items()
        )

        new._update_snr_change()

        return new

    def dropna(self, inplace=False):
        """
        Drop any channels with NaN values.

        Parameters
        ----------
        inplace : bool, optional

        Returns
        -------
        :class:`DataSet<pyproteome.data_sets.DataSet>`
        """
        new = self

        if not inplace:
            new = new.copy()

        new.psms = new.psms.dropna(
            axis=0,
            how="any",
            subset=list(new.channels.keys()),
        )

        return new

    def filter(
        self,
        ion_score_cutoff=None, confidence_cutoff=None,
        snr_cutoff=None,
        asym_snr_cutoff=None,
        p_cutoff=None,
        fold_cutoff=None,
        asym_fold_cutoff=None,
        inplace=False,
    ):
        """
        Filters a data set.

        Parameters
        ----------
        ion_score_cutoff : int, optional
        confidence_cutoff : {"High", "Medium", "Low"}, optional
        snr_cutoff : float, optional
        asym_snr_cutoff : float, optional
        p_cutoff : float, optional
        fold_cutoff : float, optional
        asym_fold_cutoff : float, optional
        inplace : bool, optional

        Returns
        -------
        :class:`DataSet<pyproteome.data_sets.DataSet>`
        """
        if ion_score_cutoff or confidence_cutoff:
            assert self.source in ["MASCOT"]

        new = self

        if not inplace:
            new = new.copy()

        confidence_levels = []

        # MASCOT confidence levels
        if confidence_cutoff:
            confidence_levels += [confidence_cutoff]
        if confidence_cutoff in ["Medium", "Low"]:
            confidence_levels += ["High"]
        if confidence_levels in ["Low"]:
            confidence_levels += ["Medium"]

        if confidence_levels:
            new.psms = new.psms[
                new.psms["Confidence Level"].isin(confidence_levels)
            ]

        # MASCOT ion scores
        if ion_score_cutoff:
            new.psms = new.psms[
                new.psms["IonScore"] >= ion_score_cutoff
            ]

        if snr_cutoff:
            new.psms = new.psms[
                abs(new.psms["SNR"]) >= snr_cutoff
            ]

        if asym_snr_cutoff:
            if asym_snr_cutoff > 0:
                new.psms = new.psms[
                    new.psms["SNR"] >= asym_snr_cutoff
                ]
            else:
                new.psms = new.psms[
                    new.psms["SNR"] <= asym_snr_cutoff
                ]

        if p_cutoff:
            new.psms = new.psms[
                new.psms["p-value"] <= p_cutoff
            ]

        if asym_fold_cutoff:
            if asym_fold_cutoff > 1:
                new.psms = new.psms[
                    new.psms["Fold Change"] >= asym_fold_cutoff
                ]
            else:
                new.psms = new.psms[
                    new.psms["Fold Change"] <= asym_fold_cutoff
                ]

        if fold_cutoff:
            if fold_cutoff < 1:
                fold_cutoff = 1 / fold_cutoff

            new.psms = new.psms[
                np.maximum.reduce(
                    [
                        new.psms["Fold Change"],
                        1 / new.psms["Fold Change"],
                    ]
                ) >= fold_cutoff
            ]

        return new

    def _update_snr_change(self):
        """
        Update a table's SNR, Fold-Change, and p-value columns.

        Values are calculated based on changes between group_a and group_b.

        Parameters
        ----------
        psms : :class:`pandas.DataFrame`
        group_a : list of str
        group_b : list of str
        normalize : bool, optional

        Returns
        -------
        :class:`pandas.DataFrame`
        """
        groups = list(self.groups.values())

        if len(groups) == 2:
            self.psms["SNR"] = pd.Series(
                [
                    _snr(row[groups[0]], row[groups[1]])
                    for index, row in self.psms.iterrows()
                ]
            )

            self.psms["Fold Change"] = pd.Series(
                self.psms[groups[0]].mean(axis=1) /
                self.psms[groups[1]].mean(axis=1)
            )
            self.psms["p-value"] = pd.Series(
                ttest_ind(
                    self.psms[groups[0]],
                    self.psms[groups[1]],
                    axis=1,
                )[1]
            )


def merge_data(
    data_sets, name=None,
    merge_duplicates=True, merge_subsets=False,
):
    """
    Merge a list of data sets together.

    Parameters
    ----------
    data_sets : list of :class:`DataSet<pyproteome.data_sets.DataSet>`
    name : str, optional
    merge_duplicates : bool, optional
    merge_subsets : bool, optional

    Returns
    -------
    :class:`DataSet<pyproteome.data_sets.DataSet>`
    """
    assert len(data_sets) > 0

    if any(not isinstance(data, DataSet) for data in data_sets):
        raise TypeError(
            "Incompatible types: {}".format(
                [type(data) for data in data_sets]
            )
        )

    if any(data.groups != data_sets[0].groups for data in data_sets):
        raise Exception("Incongruent groups between data sets.")

    new = data_sets[0].copy()
    new.psms = pd.concat([data.psms for data in data_sets])

    if merge_duplicates:
        new._merge_duplicates()

    if merge_subsets:
        new._merge_subsequences()

    new.sets = sum(data.sets for data in data_sets)
    new.enrichments = sorted(
        set(
            enrichment
            for data in data_sets
            for enrichment in data.enrichments
        )
    )
    new.tissues = sorted(
        set(
            tissue
            for data in data_sets
            for tissue in data.tissues
        )
    )

    if new.groups:
        new._update_snr_change()

    if name:
        new.name = name

    return new


def _log_cum_fold_change(vals):
    """
    Calculate the cumulative fold change (in base-2) of values.

    Fold-change is normalized to the first element in the array.

    Parameters
    ----------
    vals : :class:`numpy.ndarray` of float

    Returns
    -------
    float
    """
    return sum(
        abs(np.log2(i / (vals[0])))
        for i in vals[1:]
        if i != 0
    )


def _snr(data1, data2):
    """
    Calculate the signal-to-noise ratio between two groups of data.

    Parameters
    ----------
    data1 : :class:`numpy.ndarray` of float
    data2 : :class:`numpy.ndarray` of float

    Returns
    -------
    float
    """
    return (data1.mean() - data2.mean()) \
        / (data1.std(ddof=1) + data2.std(ddof=1))



"""Utility functions used in other modules."""

# Built-ins
import difflib
from collections import OrderedDict
import os


def fuzzy_find(needle, haystack):
    """
    Find the longest matching subsequence of needle within haystack.

    Returns the corresponding index from the beginning of needle.

    Parameters
    ----------
    needle : str
    haystack : str

    Returns
    -------
    int
    """
    s = difflib.SequenceMatcher(a=haystack, b=needle)
    best = s.find_longest_match(0, len(haystack), 0, len(needle))
    return best.a - len(needle) + best.size


def make_folder(folder_name=None):
    """
    Creates a folder if it does not exist.

    Parameters
    ----------
    folder_name: str or None
    """
    if folder_name:
        try:
            os.makedirs(folder_name)
        except OSError:
            pass


def norm(channels):
    """
    Converts a list of channels to their normalized names.

    Parameters
    ----------
    channels : list of str or OrderedDict of str, str or None

    Returns
    -------
    list of str or OrderedDict of str, str
    """
    if channels is None:
        return None

    if isinstance(channels, str):
        return channels + "_norm"

    if isinstance(channels, list):
        return [norm(i) for i in channels]

    if isinstance(channels, OrderedDict):
        return OrderedDict(
            (norm(key), val)
            for key, val in channels.items()
        )


def which(program):
    """
    Checks if a program exists in PATH's list of directories.

    Parameters
    ----------
    program : str

    Returns
    -------
    str or None
    """
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)

    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None



"""
This module provides functionality for interacting with mass spec scan data.
"""

import logging
import os
import re
import tempfile

from . import ms_labels, proteowizard


LOGGER = logging.getLogger("pycamv.scans")


class ScanQuery:
    """
    Attributes
    ----------
    scan : int
    isolation_mz : float or None
    window_offset : tuple of (int, int) or None
    precursor_scan : int or None
    collision_type : str or None
    """
    def __init__(
        self, scan,
        isolation_mz=None, window_offset=None, precursor_scan=None,
        collision_type=None,
    ):
        self.scan = scan
        self.precursor_scan = precursor_scan
        self.window_offset = window_offset
        self.isolation_mz = isolation_mz
        self.collision_type = collision_type


def _scanquery_from_spectrum(spectrum):
    """
    Parameters
    ----------
    spectrum : :class:`pymzml.spec.Spectrum<spec.Spectrum>`

    Returns
    -------
    :class:`ScanQuery<pycamv.scans.ScanQuery>`
    """
    # prefix = {"mzml": "http://psi.hupo.org/ms/mzml"}

    scan = spectrum["id"]
    isolation_mz = spectrum["isolation window target m/z"]
    window_offset = (
        spectrum["isolation window lower offset"],
        spectrum["isolation window upper offset"],
    )
    collision_type = re.search(
        ".*@([A-Za-z]+)\d+",
        spectrum["filter string"]
    ).group(1).upper()

    spectrum_ref = spectrum.xmlTreeIterFree.find(
        "precursorList/precursor",
    ).get("spectrumRef")
    precursor_scan = int(re.search("scan=(\d+)", spectrum_ref).group(1))

    return ScanQuery(
        scan,
        precursor_scan=precursor_scan,
        window_offset=window_offset,
        isolation_mz=isolation_mz,
        collision_type=collision_type,
    )


def c13_num(pep_query, scan_query):
    """
    Counts the number of C13 atoms in a query, based on the mass-error between
    the expected and isolated m/z values.

    Parameters
    ----------
    pep_query : :class:`PeptideQuery<pycamv.mascot.PeptideQuery>`
    scan_query : :class:`ScanQuery<pycamv.scans.ScanQuery>`

    Returns
    -------
    int
    """
    return int(
        round(
            pep_query.pep_exp_z *
            abs(pep_query.pep_exp_mz - scan_query.isolation_mz)
        )
    )


def get_precursor_peak_window(scan_queries, ms_data, window_size=1):
    """
    Get ion peaks around each peptide's precursor m/z range.

    Parameters
    ----------
    pep_queries : list of :class:`PeptideQuery<pycamv.mascot.PeptideQuery>`
    ms2_data : :class:`pymzml.run.Reader<run.Reader>`

    Returns
    -------
    list of list of tuple of (float, float)
    """
    def _get_percursor_peaks(query):
        window = (
            query.isolation_mz - window_size,
            query.isolation_mz + window_size,
        )

        return [
            (mz, i)
            for mz, i in ms_data[query.precursor_scan].centroidedPeaks
            if mz > window[0] and mz < window[1]
        ]

    return [
        _get_percursor_peaks(query)
        for query in scan_queries
    ]


def get_label_peak_window(pep_queries, ms2_data, window_size=1):
    """
    Get ion peaks around each peptide's label m/z range.

    Parameters
    ----------
    pep_queries : list of :class:`PeptideQuery<pycamv.mascot.PeptideQuery>`
    ms2_data : :class:`pymzml.run.Reader<run.Reader>`

    Returns
    -------
    list of list of tuple of (float, float)
    """

    def _get_labels_peaks(query):
        label_mods = query.get_label_mods

        if not label_mods:
            return []

        window = ms_labels.LABEL_MZ_WINDOW[label_mods[0]]
        window = (
            window[0] - window_size,
            window[1] + window_size
        )

        return [
            (mz, i)
            for mz, i in ms2_data[query.scan].centroidedPeaks
            if mz > window[0] and mz < window[1]
        ]

    return [
        _get_labels_peaks(pep_query)
        for pep_query in pep_queries
    ]


def get_scan_data(basename, pep_queries, out_dir=None):
    """
    Gets MS^2 and MS data for all scans in queries.

    Parameters
    ----------
    basename : str
    pep_queries : list of :class:`PeptideQuery<pycamv.mascot.PeptideQuery>`
    out_dir : str, optional

    Returns
    -------
    scan_queries : list of :class:`ScanQuery<pycamv.scans.ScanQuery>`
    ms2_data : :class:`pymzml.run.Reader<run.Reader>`
    ms_data : :class:`pymzml.run.Reader<run.Reader>`
    """
    if out_dir is None:
        out_dir = tempfile.mkdtemp()

    # Collect MS^2 data
    LOGGER.info("Converting MS^2 data.")
    ms2_data = proteowizard.raw_to_mzml(
        basename, os.path.join(out_dir, "ms2"),
        scans=sorted(set(pep_query.scan for pep_query in pep_queries)),
    )

    # Build a list of scan queries, including data about each scan
    # scan_queries = [
    #     _scanquery_from_spectrum(spectrum)
    #     for spectrum in ms2_data
    #     if spectrum["ms level"] == 2
    # ]
    scan_queries = [
        _scanquery_from_spectrum(ms2_data[pep_query.scan])
        for pep_query in pep_queries
    ]

    # Collect MS^1 data
    LOGGER.info("Converting MS^1 data.")
    ms_data = proteowizard.raw_to_mzml(
        basename, os.path.join(out_dir, "ms"),
        scans=sorted(set(i.precursor_scan for i in scan_queries)),
    )

    return scan_queries, ms2_data, ms_data



"""
This module provides functionality for normalizing protein data.

Levels can be extracted from supernatant or phosphotyrosine runs using median
or mean peptide levels across multiple channels.
"""

from __future__ import division

# Built-ins
import os
from collections import OrderedDict

# Core data analysis libraries
from matplotlib import pyplot as plt
import numpy as np

from . import utils


def get_channel_levels(
    data,
    folder_name=None, file_name=None,
):
    """
    Calculate channel levels using median channel levels.

    Should be used on sample supernatants.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`
    folder_name : str, optional
    file_name : str, optional
    """
    if folder_name is None:
        folder_name = data.name

    if not file_name:
        file_name = "channel_levels.png"

    utils.make_folder(folder_name)

    if folder_name:
        file_name = os.path.join(folder_name, file_name)

    channels = list(data.channels.keys())
    channel_levels = OrderedDict()
    base = channels[0]
    channel_levels[base] = 1

    f, axes = plt.subplots(
        len(data.channels) // 2, 2,
        sharex=True,
        figsize=(12, 12),
    )
    axes = [i for j in axes for i in j]

    for ax, col in zip(list(axes), channels[1:]):
        points = (data.psms[col] / data.psms[base]).dropna().as_matrix()

        # Filter ratios > 30, those are likely an error in quantification and
        # may throw off histogram binning
        points = points[points < 30]
        med = np.median(points)
        channel_levels[col] = med

        ax.hist(points, bins=40)
        ax.set_title(
            "{}: median: {:.2f}, $\sigma$ = {:.2f}".format(
                "{} ({})".format(data.channels[col], col)
                if isinstance(data.channels, dict) else
                col,
                med,
                points.std(ddof=1),
            )
        )
        ax.axvline(med, color='k', linestyle='--')

    for ax in axes[len(data.channels) - 1:]:
        ax.set_axis_off()

    f.show()

    if file_name:
        f.savefig(file_name)

    return channel_levels


def get_average_phospho_levels(data):
    """
    Calculate channel levels using mean channel levels.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`

    Returns
    -------
    OrderedDict
    """
    base = data.channels[0]
    levels = data.psms[list(data.channels)].mean()

    return OrderedDict(
        (name, levels[name] / levels[base])
        for name in data.channels
    )



"""
This module provides functionality for fetching protein data from UniProt.

Caches fetched protein data for faster re-use.
"""

# Built-ins
import os
import re

# Core data analysis libraries
import pandas as pd

import uniprot

from . import paths


RE_ACCESSION = re.compile("\[([A-Za-z0-9]+_[A-Z]+)\]")

UNIPROT_DATA = {}
MISS_COUNT = {}


def fetch_uniprot_data(accessions):
    """
    Fetch UniProt protein descriptions, gene names, sequences, etc.

    All information is stored in UNIPROT_DATA and can be accessed with
    get_uniprot_data().

    Parameters
    ----------
    accessions : list of str
    """
    accessions = set(accessions).difference(UNIPROT_DATA)

    if not accessions:
        return

    UNIPROT_DATA.update(
        uniprot.get_metadata_with_some_seqid_conversions(
            accessions,
            cache_dir=os.path.abspath('.uniprot_cache'),
        )
    )


def prefetch_all_uniprot():
    """
    Fetch data for all accesions found in MS Searched directory.

    Pulls all UniProt accession IDs from all "_psms.txt" files.
    """
    accessions = set()

    for filename in os.listdir(paths.MS_SEARCHED_DIR):
        if not filename.endswith("_psms.txt"):
            continue

        psms = pd.read_table(os.path.join(paths.MS_SEARCHED_DIR, filename))
        psms.dropna(
            subset=["Protein Group Accessions"],
            inplace=True,
        )
        accessions.update(
            acc.strip()
            for row_str in psms["Protein Group Accessions"]
            for acc in row_str.split(";")
        )

    fetch_uniprot_data(accessions)


def prefetch_accessions(psms):
    """
    Pre-fetch all UniProt information.

    Speeds up data processing by doing all queries in one go.

    Parameters
    ----------
    psms : :class:`pandas.DataFrame`
    """
    accessions = set()

    # MASCOT / Discoverer Outputs
    if "Protein Group Accessions" in psms.columns:
        accessions.update(
            acc.strip()
            for i in psms["Protein Group Accessions"]
            if isinstance(i, str)
            for acc in i.split(";")
        )

    # CAMV Outputs
    if "Accession" in psms.columns:
        accessions.update(
            acc.strip()
            for i in psms["Accession"]
            if isinstance(i, str)
            for acc in i.split("/")
        )

    fetch_uniprot_data(accessions)


def get_uniprot_data(accession):
    """
    Get UniProt data associated with a protein.

    Parameters
    ----------
    accession : str

    Returns
    -------
    dict
    """
    if accession not in UNIPROT_DATA:
        fetch_uniprot_data([accession])

    return UNIPROT_DATA[accession]



"""
pyproteome is a Python package for interacting with proteomics data.

It includes modules for loading, processing, and analyzing data by mass
spectometry. Currently it only supports data produced by MASCOT / Discoverer
and CAMV.
"""

from .version import __version__
from .modification import Modification, Modifications
from .sequence import Sequence, ProteinMatch
from .motif import Motif
from .protein import Protein, Proteins
from .data_sets import DataSet




import os
from pip.req import parse_requirements
from pip.download import PipSession
from setuptools import setup, find_packages

from pyproteome import __version__


REQUIREMENTS_PATH = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__), "requirements.txt",
    )
)


setup(
    name="pyproteome",
    version=__version__,
    description="Package for managing proteomics data",
    url="https://github.com/naderm/pyproteome",
    author="Nader Morshed",
    author_email="morshed@mit.edu",
    license="BSD",
    packages=find_packages(exclude=["*.tests", "tests"]),
    install_requires=[
        str(i.req)
        for i in parse_requirements(REQUIREMENTS_PATH, session=PipSession())
    ],
    dependency_links=[
        "git+https://github.com/boscoh/uniprot#egg=uniprot",
        "git+https://github.com/Phlya/adjustText#egg=adjustText"
    ],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering",
    ],
    test_suite="tests",
)



"""
This module provides functionality for interfacing with protein data.
"""

from . import fetch_data


class Proteins:
    """
    Wraps a list of proteins.

    Attributes
    ----------
    proteins : list of :class:`Protein<pyproteome.protein.Protein>`
    """

    def __init__(self, proteins):
        self.proteins = proteins

    def __iter__(self):
        return iter(self.proteins)

    def __hash__(self):
        return hash(
            tuple(self.proteins),
        )

    def __eq__(self, other):
        if isinstance(other, str):
            return (
                any(i == other for i in self.genes) or
                any(i == other for i in self.accessions)
            )

        if not isinstance(other, Proteins):
            raise TypeError(type(other))

        return len(self.proteins) == len(other.proteins) and all(
            i == j
            for i, j in zip(self.proteins, other.proteins)
        )

    def __str__(self):
        return " / ".join(
            str(i)
            for i in self.proteins
        )

    @property
    def accessions(self):
        return [i.accession for i in self.proteins]

    @property
    def genes(self):
        return [i.gene for i in self.proteins]


class Protein:
    """
    Contains information about a single protein.

    Attributes
    ----------
    accession : str
    gene : str
    description : str
    full_sequence : str
    """

    def __init__(self, accession):
        self.accession = accession

        up_data = fetch_data.get_uniprot_data(accession)

        if "gene" in up_data:
            self.gene = up_data["gene"]
        else:
            self.gene = up_data["id"]

        self.description = up_data["descriptions"][0]
        self.full_sequence = up_data["sequence"]

    def __hash__(self):
        return hash(self.accession)

    def __eq__(self, other):
        if not isinstance(other, Protein):
            raise TypeError()

        return self.accession == other.accession

    def __str__(self):
        return "{} ({})".format(
            self.description,
            self.gene,
        )



"""
This module tracks the path to user data files. Developers can override paths
here when using a custom data hierarchy.
"""

import os

BASE_DIR = os.path.abspath("..")

BCA_ASSAY_DIR = os.path.join(BASE_DIR, "BCA Protein Assays")
CAMV_OUT_DIR = os.path.join(BASE_DIR, "CAMV Output")
CAMV_SESS_DIR = os.path.join(BASE_DIR, "CAMV Sessions")
MASCOT_XML_DIR = os.path.join(BASE_DIR, "Mascot XMLs")
MS_SEARCHED_DIR = os.path.join(BASE_DIR, "MS Searched")
MS_RAW_DIR = os.path.join(BASE_DIR, "MS RAW")
SCAN_LISTS_DIR = os.path.join(BASE_DIR, "Scan Lists")
SCRIPTS_DIR = os.path.join(BASE_DIR, "Scripts")



"""
This module provides functionality for data set analysis.

Functions include volcano plots, sorted tables, and plotting sequence levels.
"""

from __future__ import division

# Built-ins
import logging
import os
import re

# IPython
from IPython.display import display

# Core data analysis libraries
from matplotlib import pyplot as plt
# import matplotlib.patches as patches
import matplotlib_venn as mv
import numpy as np
import pandas as pd
import seaborn as sns
# import scipy
from scipy.stats import ttest_ind, pearsonr
# from scipy.stats.mstats import mquantiles
# from scipy.spatial import distance
# from scipy.cluster import hierarchy
# import sklearn
# from sklearn.cluster import KMeans, MiniBatchKMeans

# Misc extras
from adjustText.adjustText import adjust_text
# import fastcluster
# import somoclu
# import uniprot

from . import fetch_data, utils


LOGGER = logging.getLogger("pyproteome.analysis")


def snr_table(
    data,
    sort="p-value",
    folder_name=None, csv_name=None,
):
    """
    Show a signal to noise table.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`
    sort : str, optional
    folder_name : str, optional
    csv_name : str, optional
    """
    if folder_name is None:
        folder_name = data.name

    if csv_name is None:
        csv_name = "{}-{}.csv".format(
            re.sub("[ ></]", "_", data.name),
            re.sub("[ ></]", "_", data.enrichment),
        )

    utils.make_folder(folder_name)

    csv_name = os.path.join(folder_name, csv_name)

    psms = data.psms[
        ["Proteins", "Sequence", "Modifications", "Fold Change", "p-value"]
    ]
    psms["Sequence"] = [
        "{} ({})".format(row["Sequence"], row["Modifications"])
        for _, row in psms.iterrows()
    ]
    if sort == "Fold Change":
        psms["Fold Change-Sort"] = psms["Fold Change"].apply(
            lambda x: max([x, 1 / x])
        )
        psms.sort_values("Fold Change-Sort", inplace=True, ascending=False)
        psms.drop("Fold Change-Sort", axis=1, inplace=True)
    else:
        psms.sort_values(sort, inplace=True, ascending=True)

    psms.drop("Modifications", axis=1, inplace=True)

    if csv_name:
        psms.to_csv(csv_name)

    return psms


def write_full_tables(datas, folder_name="All", out_name="Full Data.xlsx"):
    """
    Write a full list of data sets to a single .xlsx file.

    Parameters
    ----------
    datas : list of :class:`DataSet<pyproteome.data_sets.DataSet>`
    folder_name : str, optional
    out_name : str, optional
    """

    utils.make_folder(folder_name)

    if folder_name is not None:
        out_name = os.path.join(folder_name, out_name)

    writer = pd.ExcelWriter(out_name, engine="xlsxwriter")

    for data in datas:
        df = data.psms[["Proteins", "Sequence", "Fold Change", "p-value"]]
        df.insert(
            2, "Modifications",
            df["Sequence"].apply(
                lambda x: str(x.modifications)
            ),
        )
        df["Sequence"] = df["Sequence"].apply(str)
        df.sort_values("p-value", inplace=True, ascending=True)

        ws_name = "{}-{}".format(data.name, data.enrichment)
        df.to_excel(
            writer,
            sheet_name=ws_name,
            index=False,
        )

        ws = writer.sheets[ws_name]
        ws.freeze_panes(1, 0)
        ws.set_column(0, 0, 60)
        ws.set_column(1, 1, 30)
        ws.set_column(2, 2, 20)
        ws.set_column(3, 3, 12)
        ws.set_column(4, 4, 12)

    writer.save()


def volcano_plot(
    data,
    pval_cutoff=1.3, fold_cutoff=1.2,
    highlight=None,
    folder_name=None, title=None,
    figsize=(12, 10),
    adjust=True,
):
    """
    Display a volcano plot of data.

    This plot inclues the fold-changes and p-values associated with said
    changes.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`
    pval_cutoff : float, optional
    fold_cutoff : float, optional
    highlight : list, optional
    folder_name : str, optional
    title : str, optional
    figsize : tuple of float, float
    adjust : bool, optional
    """
    if not folder_name:
        folder_name = data.name

    utils.make_folder(folder_name)

    if not title:
        title = "{} - {} - (Bio-N={}, Tech-N={})".format(
            data.tissue,
            data.enrichment,
            min(len(group) for group in data.groups.values()),
            data.sets,
        )
        if abs(fold_cutoff - 1.2) > 0.1:
            title += " -- Fold Change > {}".format(fold_cutoff)

    if title:
        file_name = re.sub("[ ></]", "_", title) + "_Volcano.png"

        if folder_name:
            file_name = os.path.join(folder_name, file_name)

    upper_fold = np.log2(fold_cutoff)
    lower_fold = -upper_fold

    # Calculate the Fold-Change / p-values
    pvals = []
    changes = []
    sig_pvals = []
    sig_changes = []
    sig_labels = []
    colors = []

    for index, row in data.psms.iterrows():
        color = "grey"
        row_pval = -np.log10(row["p-value"])
        row_change = np.log2(row["Fold Change"])

        pvals.append(row_pval)
        changes.append(row_change)

        if row_pval > pval_cutoff and \
           (row_change > upper_fold or row_change < lower_fold):
            row_label = " / ".join(row["Proteins"].genes)

            sig_pvals.append(row_pval)
            sig_changes.append(row_change)
            sig_labels.append(row_label)
            color = "blue"

        colors.append(color)

    # Draw the figure
    fig, ax = plt.subplots(figsize=figsize)
    ax.scatter(changes, pvals, c=colors)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.set_xlabel("$log_2$ Fold Change")
    ax.set_ylabel("$-log_{10}$ p-value")
    ax.axhline(pval_cutoff, color="r", linestyle="dashed", linewidth=0.5)
    ax.axvline(upper_fold, color="r", linestyle="dashed", linewidth=0.5)
    ax.axvline(lower_fold, color="r", linestyle="dashed", linewidth=0.5)
    ax.set_ylim(bottom=-0.1)

    # Position the labels
    texts = []
    for x, y, txt in zip(sig_changes, sig_pvals, sig_labels):
        text = ax.text(x, y, txt)

        if highlight and txt in highlight:
            text.set_fontsize(20)

        text.set_bbox(
            dict(
                color="lightgreen" if x > 0 else "pink",
                alpha=0.8,
                edgecolor="red",
            )
        )

        texts.append(text)

    if adjust:
        adjust_text(
            x=sig_changes,
            y=sig_pvals,
            texts=texts,
            ax=ax,
            lim=400,
            force_text=0.1,
            force_points=0.1,
            arrowprops=dict(arrowstyle="->", relpos=(0, 0), lw=1),
            only_move={
                "points": "y",
                "text": "xy",
            }
        )
    else:
        for j, text in enumerate(texts):
            a = ax.annotate(
                text.get_text(),
                xy=text.get_position(),
                xytext=text.get_position(),
                arrowprops=dict(arrowstyle="->", relpos=(0, 0), lw=1),
            )
            a.__dict__.update(text.__dict__)
            a.draggable()
            texts[j].remove()

    if title:
        ax.set_title(title)
        fig.savefig(
            file_name,
            bbox_inches="tight", dpi=500,
            transparent=True,
        )

    fig.show()


def hierarchical_heatmap(
    data,
    baseline_channels=None,
    metric="euclidean", method="centroid",
):
    """
    Plot a hierarhically-clustered heatmap of a data set.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`
    baseline_channels : list of str, optional
        List of channels to average and use as baseline for each row.
    metric : str, optional
        Hierarchical clustering distance metric.
    method : str, optional
        Hierarchical clustering method.
    """
    channels = [
        channel
        for channels in data.groups.values()
        for channel in channels
    ]
    channel_names = [
        data.channels[i]
        for i in channels
    ]

    if baseline_channels is None:
        baseline_channels = list(data.groups.values())[-1]

    raw = data.psms[channels].as_matrix()
    raw_baseline = data.psms[baseline_channels].as_matrix().mean(axis=1)
    raw = np.log2((raw.T / raw_baseline).T)

    sns.clustermap(
        raw,
        method=method,
        metric=metric,
        col_cluster=False,
        xticklabels=channel_names,
        yticklabels=False,
    )


def venn3(data_a, data_b, data_c, folder_name=None, filename=None):
    """
    Display a three-way venn diagram between data set sequences.

    Parameters
    ----------
    data_a : :class:`DataSet<pyproteome.data_sets.DataSet>`
    data_b : :class:`DataSet<pyproteome.data_sets.DataSet>`
    data_c : :class:`DataSet<pyproteome.data_sets.DataSet>`
    folder_name : str, optional
    filename : str, optional
    """
    utils.make_folder(folder_name)

    if folder_name and filename:
        filename = os.path.join(folder_name, filename)

    group_a = set(data_a["Sequence"])
    group_b = set(data_b["Sequence"])
    group_c = set(data_c["Sequence"])

    f = plt.figure(figsize=(12, 12))
    v = mv.venn3(
        subsets=(
            len(group_a.difference(group_b).difference(group_c)),
            len(group_b.difference(group_a).difference(group_c)),
            len(group_a.intersection(group_b).difference(group_c)),
            len(group_c.difference(group_a).difference(group_b)),
            len(group_a.intersection(group_c).difference(group_b)),
            len(group_b.intersection(group_c).difference(group_a)),
            len(group_a.intersection(group_b).intersection(group_c)),
        ),
        set_labels=(data_a.tissue, data_b.tissue, data_c.tissue),
    )

    for label in v.set_labels:
        if label:
            label.set_fontsize(32)

    for label in v.subset_labels:
        if label:
            label.set_fontsize(20)

    f.show()

    if filename:
        f.savefig(filename, transparent=True)


def write_lists(
    data,
    folder_name=None, sorted_name="sorted_list.txt",
    hits_name="hits_list.txt", background_name="back_list.txt",
):
    """
    Write a list of peptides to files.

    Includes peptides sorted by fold change, significantly changing peptides,
    and background peptides with little change.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`
    folder_name : str, optional
    sorted_name : str, optional
    hits_name : str, optional
    background_name : str, optional
    """
    if folder_name is None:
        folder_name = data.name

    utils.make_folder(folder_name)

    if folder_name:
        sorted_name = os.path.join(folder_name, sorted_name)
        hits_name = os.path.join(folder_name, hits_name)
        background_name = os.path.join(folder_name, background_name)

    change_psms = data.psms.copy()
    change_psms["Fold Change"] = np.maximum.reduce(
        [
            change_psms["Fold Change"],
            1 / change_psms["Fold Change"],
        ]
    )

    with open(sorted_name, "w") as f:
        f.write(
            "\n".join(
                i.accessions[0]
                for i in change_psms.sort(
                    "Fold Change",
                    ascending=False,
                )["Proteins"].drop_duplicates(keep="first")
            )
        )
    with open(hits_name, "w") as f:
        f.write(
            "\n".join(
                i.accessions[0]
                for i in data.filter(
                    fold_cutoff=1.3,
                    snr_cutoff=1,
                ).psms["Proteins"].drop_duplicates(keep="first")
            )
        )

    with open(background_name, "w") as f:
        f.write(
            "\n".join(
                i.accessions[0]
                for i in data.psms["Proteins"].drop_duplicates(keep="first")
            )
        )


def plot_sequence_between(
    data, sequences,
):
    """
    Plot the levels of a sequence between two groups.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`
    sequences : list of str
    """
    groups = list(reversed(data.groups.values()))
    labels = list(reversed(data.groups.keys()))

    psms = data.psms.copy()
    psms["Seq Str"] = psms["Sequence"].apply(str)
    psms = psms[psms["Seq Str"].isin(sequences)]

    points = np.array(
        [
            psms[group].as_matrix().sum(axis=0)
            for group in groups
        ]
    )

    values = points.mean(axis=1)
    errs = points.std(axis=1)

    f, ax = plt.subplots()

    indices = np.arange(len(values))
    bar_width = .35
    ax.bar(
        bar_width + indices,
        values,
        bar_width,
        yerr=errs,
        ecolor="k",
    )

    ax.set_ylabel(
        "Cumulative Channel Signal{}".format(
            " (Normalized)" if data.normalized else ""
        ),
        fontsize=20,
    )
    ax.ticklabel_format(style="plain")

    for label in ax.get_yticklabels():
        label.set_fontsize(14)

    ax.set_xticks(indices + bar_width * 1.5)
    ax.set_xticklabels(labels, fontsize=16)

    title = "{}".format(
        " / ".join(sequences),
    )
    ax.set_title(title, fontsize=20)
    ax.xaxis.grid(False)

    def _wrap_list(val):
        if isinstance(val, float):
            return [val]
        return val

    # XXX: Not correct...
    display(
        dict(
            zip(
                sequences,
                _wrap_list(
                    ttest_ind(
                        points[0, :].T,
                        points[1, :].T
                    )[1]
                )
            )
        )
    )

    return f


def plot_sequence(
    data, sequence,
):
    """
    Plot the levels of a sequence across multiple channels.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`
    sequence : str or :class:`Sequence<pyproteome.sequence.Sequence>`
    """
    channels = list(data.channels.keys())

    psms = data.psms[data.psms["Sequence"] == sequence]

    values = psms[channels].as_matrix()
    values = (values.T / values[:, 0]).T

    f, ax = plt.subplots()

    for i in range(values.shape[0]):
        indices = np.arange(len(values[i]))
        bar_width = .35
        ax.bar(bar_width + indices, values[i], bar_width)
        ax.set_xticks(indices + bar_width * 1.5)
        ax.set_xticklabels(list(data.channels.values()))

    display(values)


def plot_correlation(
    data1, data2, folder_name=None, filename=None,
    adjust=True, label_cutoff=1.5,
):
    """
    Plot the correlation between peptides levels in two different data sets.

    Parameters
    ----------
    data1 : :class:`DataSet<pyproteome.data_sets.DataSet>`
    data2 : :class:`DataSet<pyproteome.data_sets.DataSet>`
    folder_name : str, optional
    filename : str, optional
    """
    if not folder_name:
        folder_name = "All" if data1.name != data2.name else data1.name

    utils.make_folder(folder_name)

    if folder_name and filename:
        filename = os.path.join(folder_name, filename)

    merged = pd.merge(
        data1.psms, data2.psms,
        on="Sequence",
    ).dropna()

    f, ax = plt.subplots()
    ax.scatter(
        np.log2(merged["Fold Change_x"]),
        np.log2(merged["Fold Change_y"]),
    )

    label_cutoff = np.log2(label_cutoff)
    texts, x_s, y_s = [], [], []

    for index, row in merged.iterrows():
        x = row["Fold Change_x"]
        y = row["Fold Change_y"]
        ratio = np.log2(x / y)
        x, y = np.log2(x), np.log2(y)

        if ratio < label_cutoff and ratio > - label_cutoff:
            continue

        x_s.append(x)
        y_s.append(y)

        text = ax.text(
            x, y, " / ".join(row["Proteins_x"].genes),
        )

        text.set_bbox(
            dict(
                color="lightgreen" if ratio < 0 else "pink",
                alpha=0.8,
                edgecolor="red",
            )
        )
        texts.append(text)

    if adjust:
        adjust_text(
            x=x_s,
            y=y_s,
            texts=texts,
            ax=ax,
            lim=400,
            force_text=0.1,
            force_points=0.1,
            arrowprops=dict(arrowstyle="->", relpos=(0, 0), lw=1),
            only_move={
                "points": "y",
                "text": "xy",
            }
        )
    else:
        for j, text in enumerate(texts):
            a = ax.annotate(
                text.get_text(),
                xy=text.get_position(),
                xytext=text.get_position(),
                arrowprops=dict(arrowstyle="->", relpos=(0, 0), lw=1),
            )
            a.__dict__.update(text.__dict__)
            a.draggable()
            texts[j].remove()

    min_x = min(np.log2(merged["Fold Change_x"]))
    max_x = max(np.log2(merged["Fold Change_x"]))
    ax.plot([min_x, max_x], [min_x, max_x], "--")
    ax.plot(
        [min_x, max_x],
        [min_x + label_cutoff, max_x + label_cutoff],
        color="lightgreen",
        linestyle=":",
    )
    ax.plot(
        [min_x, max_x],
        [min_x - label_cutoff, max_x - label_cutoff],
        color="pink",
        linestyle=":",
    )

    ax.set_xlabel("$log_2$ Fold Change -- {}".format(data1.name))
    ax.set_ylabel("$log_2$ Fold Change -- {}".format(data2.name))

    corr = pearsonr(merged["Fold Change_x"], merged["Fold Change_y"])[0]

    ax.set_title(
        "Correlation: {:.2f}".format(corr)
    )

    if filename:
        f.savefig(filename, transparent=True)


def find_tfs(data, folder_name=None, csv_name=None):
    """
    Scan over a data set to find proteins annotated as transcription factors.

    Parameters
    ----------
    data : :class:`DataSet<pyproteome.data_sets.DataSet>`
    folder_name : str, optional
    csv_name : str, optional
    """
    if folder_name is None:
        folder_name = data.name

    utils.make_folder(folder_name)

    if csv_name is None:
        csv_name = "Changing TFs.csv"

    if folder_name and csv_name:
        csv_name = os.path.join(folder_name, csv_name)

    def _is_tf(prots):
        go_terms = (
            "DNA binding",
            "double-stranded DNA binding",
            "transcription factor binding",
        )
        return any(
            go_term in go
            for prot in prots
            for go in fetch_data.get_uniprot_data(prot.accession).get(
                "go", [],
            )
            for go_term in go_terms
        )

    tfs = data.psms[data.psms["Proteins"].apply(_is_tf)]
    tfs.sort(columns="Fold Change", ascending=False, inplace=True)

    if csv_name:
        tfs[["Proteins", "Sequence", "Modifications", "Fold Change"]].to_csv(
            csv_name,
            index=False,
        )

    return tfs



"""
This module provides functionality for interacting with CAMV.

Currently limited to importing and outputing scan lists.
"""

# Built-ins
from __future__ import division

from collections import OrderedDict
import logging
import re
import tempfile
import shutil

# Core data analysis libraries

from . import compare, fragments, gen_sequences, mascot, ms_labels, scans


LOGGER = logging.getLogger("pycamv.validate")


class SearchOptions:
    """
    Contains options used to search a data set in MASCOT.

    Attributes
    ----------
    label_type : tuple of (str, int)
    """
    def __init__(self, fixed_mods, var_mods):
        self.label_type = (None, 0)

        for mod in fixed_mods:
            mod = re.sub(r"([\w-]+) \([\w-]+\)", r"\1", mod)
            num = ms_labels.LABEL_NUMBERS.get(mod, 0)

            if num > 0:
                self.label_type = ("Fixed", num)

        for mod in var_mods:
            mod = re.sub(r"([\w-]+) \([\w-]+\)", r"\1", mod)
            num = ms_labels.LABEL_NUMBERS.get(mod, 0)

            if num > 0:
                self.label_type = ("Variable", num)

        # TODO: Parse out SILAC, C-mod, phospho, etc


def validate_spectra(basename, scan_list=None):
    """
    Generate CAMV web page for validating spectra.

    Parameters
    ----------
    basename : str
    scan_list : list of int, optional
    """
    # Read MASCOT xml file
    xml_name = "{}.xml".format(basename)

    fixed_mods, var_mods, pep_queries = mascot.read_mascot_xml(xml_name)

    # Optionally filter queries using a scan list
    if scan_list is not None:
        pep_queries = [
            query
            for query in pep_queries
            if query.scan in scan_list
        ]

    # Extract MASCOT search options
    options = SearchOptions(fixed_mods, var_mods)

    # Remove peptides with an excess of modification combinations

    # Get scan data from RAW file
    out_dir = tempfile.mkdtemp()

    LOGGER.info("Getting scan data.")
    scan_queries, ms_two_data, ms_data = scans.get_scan_data(
        basename, pep_queries, out_dir,
    )

    scan_mapping = OrderedDict(
        (pep_query, scan_query)
        for pep_query, scan_query in zip(pep_queries, scan_queries)
    )

    # Generate sequences
    LOGGER.info("Generating sequences.")
    sequence_mapping = OrderedDict(
        (
            pep_query,
            tuple(
                gen_sequences.gen_possible_seq(
                    pep_query.pep_seq,
                    pep_query.pep_var_mods,
                )
            ),
        )
        for pep_query in pep_queries
    )

    LOGGER.info("Generating fragment ions.")
    fragment_mapping = OrderedDict(
        (
            (pep_query, tuple(sequence)),
            fragments.fragment_ions(
                sequence, pep_query.pep_exp_z,
            ),
        )
        for pep_query, sequences in sequence_mapping.items()
        for sequence in sequences
    )

    LOGGER.info("Comparing predicted peaks to spectra.")
    peak_hits = {
        (pep_query, sequence): compare.compare_spectra(
            ms_two_data[pep_query.scan],
            frag_ions,
            pep_query.pep_exp_z,
            scans.c13_num(pep_query, scan_mapping[pep_query]),
            tol=compare.COLLISION_TOLS[scan_mapping[pep_query].collision_type],
        )
        for (pep_query, sequence), frag_ions in list(fragment_mapping.items())
    }

    # XXX: Determine SILAC precursor masses?

    LOGGER.info("Collecting precursor ion peaks.")
    precursor_windows = OrderedDict(
        zip(
            pep_queries,
            scans.get_precursor_peak_window(scan_queries, ms_data)
        )
    )

    LOGGER.info("Collecting peptide label peaks.")
    label_windows = OrderedDict(
        zip(
            pep_queries,
            scans.get_label_peak_window(pep_queries, ms_two_data)
        )
    )

    # XXX: Remove precursor contaminated scans from validation list?

    # Check each assignment to each scan

    # Output data

    shutil.rmtree(out_dir)

    return options, peak_hits, precursor_windows, label_windows
